from pyspark.sql import Column
from pyspark.sql.functions import filter, size, transform, when, lit, array


def remove_empty_maps(column: Column) -> Column:
    """
    This function takes a column of type array(map(str,str)) and removes empty maps from it
    Args:
        column: Provide a column of type array(map(str,str))
        Returns:
                list: Returns a Column which is not having empty maps
    """
    # The below line is already tested in test_udf.py but not shown in coverage. So ignoring it for now!
    return filter(column, lambda x: size(x) > 0)  # pragma: no cover


def get_actions_list(column: Column) -> Column:
    """
    This function takes column of type array(map(str,str)) and creates list by picking action_if_failed from dict
    Args:
        column: Provide a column of type array(map(str,str))

    Returns:
           list: returns a column with list of action_if_failed from the set expectations rules

    """

    action_if_failed = transform(column, lambda x: x["action_if_failed"])
    return when(size(action_if_failed) == 0, array(lit("ignore"))).otherwise(
        action_if_failed
    )  # pragma: no cover



import functools
import pluggy
from spark_expectations import _log
from spark_expectations.notifications.plugins.base_notification import (
    SparkExpectationsNotification,
    SPARK_EXPECTATIONS_NOTIFICATION_PLUGIN,
)

from spark_expectations.notifications.plugins.email import (
    SparkExpectationsEmailPluginImpl,
)
from spark_expectations.notifications.plugins.slack import (
    SparkExpectationsSlackPluginImpl,
)
from spark_expectations.notifications.plugins.teams import (
    SparkExpectationsTeamsPluginImpl,
)


@functools.lru_cache
def get_notifications_hook() -> pluggy.PluginManager:
    """
    function provides pluggy hook manger to send email and slack notification
    Returns:
        PluginManager: pluggy Manager object

    """
    pm = pluggy.PluginManager(SPARK_EXPECTATIONS_NOTIFICATION_PLUGIN)
    pm.add_hookspecs(SparkExpectationsNotification)
    pm.register(
        SparkExpectationsEmailPluginImpl(), "spark_expectations_email_notification"
    )
    pm.register(
        SparkExpectationsSlackPluginImpl(), "spark_expectations_slack_notification"
    )
    pm.register(
        SparkExpectationsTeamsPluginImpl(), "spark_expectations_teams_notification"
    )
    for name, plugin_instance in pm.list_name_plugin():
        _log.info(
            "Loaded plugin with name: %s and class: %s",
            name,
            plugin_instance.__class__.__name__,
        )
    return pm


_notification_hook = get_notifications_hook().hook



from dataclasses import dataclass
from typing import Dict, List, Tuple, Optional, Any
from pyspark.sql import DataFrame
from spark_expectations.core.exceptions import (
    SparkExpectationsMiscException,
)
from spark_expectations import _log
from spark_expectations.notifications.push.spark_expectations_notify import (
    SparkExpectationsNotify,
)
from spark_expectations.core.context import SparkExpectationsContext
from spark_expectations.sinks.utils.writer import SparkExpectationsWriter
from spark_expectations.utils.actions import SparkExpectationsActions


@dataclass
class SparkExpectationsRegulateFlow:
    """
    This is helper class and implements/supports running data quality flow
    """

    product_id: str

    @staticmethod
    def execute_dq_process(
        _context: SparkExpectationsContext,
        _actions: SparkExpectationsActions,
        _writer: SparkExpectationsWriter,
        _notification: SparkExpectationsNotify,
        expectations: Dict[str, List[dict]],
        table_name: str,
        _input_count: int = 0,
    ) -> Any:
        """
        This functions takes required static variable and returns the function
        Args:
            _context: SparkExpectationsContext class object
            _actions: SparkExpectationsActions class object
            _writer: SparkExpectationsWriter class object
            _notification: SparkExpectationsNotify class object
            expectations: expectations dictionary which contains rules
            table_name: name of the table
            _input_count: number of records in the source dataframe
        Returns:
               Any: returns function

        """

        def func_process(
            df: DataFrame,
            _rule_type: str,
            row_dq_flag: bool = False,
            source_agg_dq_flag: bool = False,
            final_agg_dq_flag: bool = False,
            source_query_dq_flag: bool = False,
            final_query_dq_flag: bool = False,
            error_count: int = 0,
            output_count: int = 0,
        ) -> Tuple[DataFrame, Optional[List[Dict[str, str]]], int, str]:
            """
            This inner function helps to process data quality rules based on different rules types
            Args:
                df: dataframe for data quality
                _rule_type: type of the rule
                row_dq_flag: default false, Mark True tp process row level data quality
                source_agg_dq_flag: default false, Mark True tp process agg level data quality on source dataframe
                final_agg_dq_flag: default false, Mark True tp process agg level data quality on final dataframe
                source_query_dq_flag: default false, Mark True tp process query level data quality on source dataframe
                final_query_dq_flag: default false, Mark True tp process query level data quality on final dataframe
                error_count: number of records error records (default zero)
                output_count: number of output records from expectations (default zero)

            Returns:
                   Tuples with data frame which contains dq result, agg result in list, error count and
                   status of the flow

            """
            try:
                _error_df: Optional[DataFrame] = None
                _error_count: int = error_count

                _running_rule_type_name = (
                    _context.get_row_dq_rule_type_name
                    if row_dq_flag
                    else (
                        _context.get_agg_dq_rule_type_name
                        if (source_agg_dq_flag or final_agg_dq_flag)
                        else _context.get_query_dq_rule_type_name
                    )
                )

                _log.info(
                    "The data quality dataframe is getting created for expectations"
                )

                _df_dq: DataFrame = _actions.run_dq_rules(
                    _context,
                    df,
                    expectations,
                    _running_rule_type_name,
                    _source_dq_enabled=(
                        source_query_dq_flag is True or source_agg_dq_flag is True
                    ),
                    _target_dq_enabled=(
                        final_query_dq_flag is True or final_agg_dq_flag is True
                    ),
                )

                _log.info("The data quality dataframe is created for expectations")
                _context.print_dataframe_with_debugger(_df_dq)

                agg_dq_res = (
                    _actions.create_agg_dq_results(
                        _context, _df_dq, _running_rule_type_name
                    )
                    if row_dq_flag is False
                    else None
                )

                if row_dq_flag:
                    _log.info("Writing error records into the table started")

                    _error_count, _error_df = _writer.write_error_records_final(
                        _df_dq,
                        f"{table_name}_error",
                        _context.get_row_dq_rule_type_name,
                    )
                    if _context.get_summarised_row_dq_res:
                        _notification.notify_rules_exceeds_threshold(expectations)
                        _writer.generate_rules_exceeds_threshold(expectations)

                    _context.print_dataframe_with_debugger(_error_df)

                    # set the error count
                    _context.set_error_count(_error_count)

                # set agg result
                if source_agg_dq_flag:
                    _context.set_source_agg_dq_result(agg_dq_res)
                elif final_agg_dq_flag:
                    _context.set_final_agg_dq_result(agg_dq_res)
                elif source_query_dq_flag:
                    _context.set_source_query_dq_result(agg_dq_res)
                elif final_query_dq_flag:
                    _context.set_final_query_dq_result(agg_dq_res)

                df = _actions.action_on_rules(
                    _context,
                    _error_df if row_dq_flag else _df_dq,
                    _input_count,
                    _error_count=_error_count,
                    _output_count=output_count,
                    _rule_type=_running_rule_type_name,
                    _row_dq_flag=row_dq_flag,
                    _source_agg_dq_flag=source_agg_dq_flag,
                    _final_agg_dq_flag=final_agg_dq_flag,
                    _source_query_dq_flag=source_query_dq_flag,
                    _final_query_dq_flag=final_query_dq_flag,
                )
                _context.print_dataframe_with_debugger(df)

                return df, agg_dq_res, _error_count, "Passed"

            except Exception as e:
                raise SparkExpectationsMiscException(
                    f"error occurred while executing func_process {e}"
                )

        return func_process



from dataclasses import dataclass
from typing import Dict, Optional, Tuple, List
from datetime import datetime
from pyspark.sql import DataFrame
from pyspark.sql.functions import (
    lit,
    expr,
    when,
    array,
    to_timestamp,
    round as sql_round,
    create_map,
    explode,
    to_json,
    col,
)
from spark_expectations import _log
from spark_expectations.core.exceptions import (
    SparkExpectationsUserInputOrConfigInvalidException,
    SparkExpectationsMiscException,
)
from spark_expectations.secrets import SparkExpectationsSecretsBackend
from spark_expectations.utils.udf import remove_empty_maps
from spark_expectations.core.context import SparkExpectationsContext
from spark_expectations.sinks import _sink_hook
from spark_expectations.config.user_config import Constants as user_config


@dataclass
class SparkExpectationsWriter:
    """
    This class implements/supports writing data into the sink system
    """

    _context: SparkExpectationsContext

    def __post_init__(self) -> None:
        self.spark = self._context.spark

    def save_df_as_table(
        self, df: DataFrame, table_name: str, config: dict, stats_table: bool = False
    ) -> None:
        """
        This function takes a dataframe and writes into a table

        Args:
            df: Provide the dataframe which need to be written as a table
            table_name: Provide the table name to which the dataframe need to be written to
            config: Provide the config to write the dataframe into the table
            stats_table: Provide if this is for writing stats table

        Returns:
            None:

        """
        try:
            print("run date ", self._context.get_run_date)
            if not stats_table:
                df = df.withColumn(
                    self._context.get_run_id_name, lit(f"{self._context.get_run_id}")
                ).withColumn(
                    self._context.get_run_date_name,
                    to_timestamp(
                        lit(f"{self._context.get_run_date}"), "yyyy-MM-dd HH:mm:ss"
                    ),
                )
            _log.info("_save_df_as_table started")

            _df_writer = df.write

            if config["mode"] is not None:
                _df_writer = _df_writer.mode(config["mode"])
            if config["format"] is not None:
                _df_writer = _df_writer.format(config["format"])
            if config["partitionBy"] is not None and config["partitionBy"] != []:
                _df_writer = _df_writer.partitionBy(config["partitionBy"])
            if config["sortBy"] is not None and config["sortBy"] != []:
                _df_writer = _df_writer.sortBy(config["sortBy"])
            if config["bucketBy"] is not None and config["bucketBy"] != {}:
                bucket_by_config = config["bucketBy"]
                _df_writer = _df_writer.bucketBy(
                    bucket_by_config["numBuckets"], bucket_by_config["colName"]
                )
            if config["options"] is not None and config["options"] != {}:
                _df_writer = _df_writer.options(**config["options"])

            _log.info("Writing records to table: %s", table_name)

            if config["format"] == "bigquery":
                _df_writer.option("table", table_name).save()
            else:
                _df_writer.saveAsTable(name=table_name)
                _log.info("finished writing records to table: %s,", table_name)
                if not stats_table:
                    # Fetch table properties
                    table_properties = self.spark.sql(
                        f"SHOW TBLPROPERTIES {table_name}"
                    ).collect()
                    table_properties_dict = {
                        row["key"]: row["value"] for row in table_properties
                    }

                    # Set product_id in table properties
                    if (
                        table_properties_dict.get("product_id") is None
                        or table_properties_dict.get("product_id")
                        != self._context.product_id
                    ):
                        _log.info(
                            "product_id is not set for table %s in tableproperties, setting it now",
                            table_name,
                        )
                        self.spark.sql(
                            f"ALTER TABLE {table_name} SET TBLPROPERTIES ('product_id' = "
                            f"'{self._context.product_id}')"
                        )

        except Exception as e:
            raise SparkExpectationsUserInputOrConfigInvalidException(
                f"error occurred while writing data in to the table - {table_name}: {e}"
            )

    def write_error_stats(self) -> None:
        """
        This functions takes the stats table and write it into error table

        Args:
            config: Provide the config to write the dataframe into the table

        Returns:
            None:

        """
        try:
            self.spark.conf.set("spark.sql.session.timeZone", "Etc/UTC")
            from datetime import date

            table_name: str = self._context.get_table_name
            input_count: int = self._context.get_input_count
            error_count: int = self._context.get_error_count
            output_count: int = self._context.get_output_count
            source_agg_dq_result: Optional[
                List[Dict[str, str]]
            ] = self._context.get_source_agg_dq_result
            final_agg_dq_result: Optional[
                List[Dict[str, str]]
            ] = self._context.get_final_agg_dq_result
            source_query_dq_result: Optional[
                List[Dict[str, str]]
            ] = self._context.get_source_query_dq_result
            final_query_dq_result: Optional[
                List[Dict[str, str]]
            ] = self._context.get_final_query_dq_result

            error_stats_data = [
                (
                    self._context.product_id,
                    table_name,
                    input_count,
                    error_count,
                    output_count,
                    self._context.get_output_percentage,
                    self._context.get_success_percentage,
                    self._context.get_error_percentage,
                    source_agg_dq_result
                    if source_agg_dq_result and len(source_agg_dq_result) > 0
                    else None,
                    final_agg_dq_result
                    if final_agg_dq_result and len(final_agg_dq_result) > 0
                    else None,
                    source_query_dq_result
                    if source_query_dq_result and len(source_query_dq_result) > 0
                    else None,
                    final_query_dq_result
                    if final_query_dq_result and len(final_query_dq_result) > 0
                    else None,
                    self._context.get_summarised_row_dq_res,
                    self._context.get_rules_exceeds_threshold,
                    {
                        "run_status": self._context.get_dq_run_status,
                        "source_agg_dq": self._context.get_source_agg_dq_status,
                        "source_query_dq": self._context.get_source_query_dq_status,
                        "row_dq": self._context.get_row_dq_status,
                        "final_agg_dq": self._context.get_final_agg_dq_status,
                        "final_query_dq": self._context.get_final_query_dq_status,
                    },
                    {
                        "run_time": self._context.get_dq_run_time,
                        "source_agg_dq_run_time": self._context.get_source_agg_dq_run_time,
                        "source_query_dq_run_time": self._context.get_source_query_dq_run_time,
                        "row_dq_run_time": self._context.get_row_dq_run_time,
                        "final_agg_dq_run_time": self._context.get_final_agg_dq_run_time,
                        "final_query_dq_run_time": self._context.get_final_query_dq_run_time,
                    },
                    {
                        "rules": {
                            "num_row_dq_rules": self._context.get_num_row_dq_rules,
                            "num_dq_rules": self._context.get_num_dq_rules,
                        },
                        "agg_dq_rules": self._context.get_num_agg_dq_rules,
                        "query_dq_rules": self._context.get_num_query_dq_rules,
                    },
                    self._context.get_run_id,
                    date.fromisoformat(self._context.get_run_date[0:10]),
                    datetime.strptime(
                        self._context.get_run_date,
                        "%Y-%m-%d %H:%M:%S",
                    ),
                )
            ]
            error_stats_rdd = self.spark.sparkContext.parallelize(error_stats_data)

            from pyspark.sql.types import (
                StructType,
                StructField,
                StringType,
                IntegerType,
                LongType,
                FloatType,
                DateType,
                ArrayType,
                MapType,
                TimestampType,
            )

            error_stats_schema = StructType(
                [
                    StructField("product_id", StringType(), True),
                    StructField("table_name", StringType(), True),
                    StructField("input_count", LongType(), True),
                    StructField("error_count", LongType(), True),
                    StructField("output_count", LongType(), True),
                    StructField("output_percentage", FloatType(), True),
                    StructField("success_percentage", FloatType(), True),
                    StructField("error_percentage", FloatType(), True),
                    StructField(
                        "source_agg_dq_results",
                        ArrayType(MapType(StringType(), StringType())),
                        True,
                    ),
                    StructField(
                        "final_agg_dq_results",
                        ArrayType(MapType(StringType(), StringType())),
                        True,
                    ),
                    StructField(
                        "source_query_dq_results",
                        ArrayType(MapType(StringType(), StringType())),
                        True,
                    ),
                    StructField(
                        "final_query_dq_results",
                        ArrayType(MapType(StringType(), StringType())),
                        True,
                    ),
                    StructField(
                        "row_dq_res_summary",
                        ArrayType(MapType(StringType(), StringType())),
                        True,
                    ),
                    StructField(
                        "row_dq_error_threshold",
                        ArrayType(MapType(StringType(), StringType())),
                        True,
                    ),
                    StructField("dq_status", MapType(StringType(), StringType()), True),
                    StructField(
                        "dq_run_time", MapType(StringType(), FloatType()), True
                    ),
                    StructField(
                        "dq_rules",
                        MapType(StringType(), MapType(StringType(), IntegerType())),
                        True,
                    ),
                    StructField(self._context.get_run_id_name, StringType(), True),
                    StructField(self._context.get_run_date_name, DateType(), True),
                    StructField(
                        self._context.get_run_date_time_name, TimestampType(), True
                    ),
                ]
            )

            df = self.spark.createDataFrame(error_stats_rdd, schema=error_stats_schema)
            self._context.print_dataframe_with_debugger(df)

            df = (
                df.withColumn("output_percentage", sql_round(df.output_percentage, 2))
                .withColumn("success_percentage", sql_round(df.success_percentage, 2))
                .withColumn("error_percentage", sql_round(df.error_percentage, 2))
            )
            _log.info(
                "Writing metrics to the stats table: %s, started",
                self._context.get_dq_stats_table_name,
            )
            if self._context.get_stats_table_writer_config["format"] == "bigquery":
                df = df.withColumn("dq_rules", to_json(df["dq_rules"]))

            self.save_df_as_table(
                df,
                self._context.get_dq_stats_table_name,
                config=self._context.get_stats_table_writer_config,
                stats_table=True,
            )

            _log.info(
                "Writing metrics to the stats table: %s, ended",
                {self._context.get_dq_stats_table_name},
            )

            # TODO check if streaming_stats is set to off, if it's enabled only then this should run

            _se_stats_dict = self._context.get_se_streaming_stats_dict
            if _se_stats_dict["se.enable.streaming"]:
                secret_handler = SparkExpectationsSecretsBackend(
                    secret_dict=_se_stats_dict
                )
                kafka_write_options: dict = (
                    {
                        "kafka.bootstrap.servers": "localhost:9092",
                        "topic": self._context.get_se_streaming_stats_topic_name,
                        "failOnDataLoss": "true",
                    }
                    if self._context.get_env == "local"
                    else (
                        {
                            "kafka.bootstrap.servers": f"{secret_handler.get_secret(self._context.get_server_url_key)}",
                            "kafka.security.protocol": "SASL_SSL",
                            "kafka.sasl.mechanism": "OAUTHBEARER",
                            "kafka.sasl.jaas.config": "kafkashaded.org.apache.kafka.common.security.oauthbearer."
                            "OAuthBearerLoginModule required oauth.client.id="
                            f"'{secret_handler.get_secret(self._context.get_client_id)}'  "
                            + "oauth.client.secret="
                            f"'{secret_handler.get_secret(self._context.get_token)}' "
                            "oauth.token.endpoint.uri="
                            f"'{secret_handler.get_secret(self._context.get_token_endpoint_url)}'; ",
                            "kafka.sasl.login.callback.handler.class": "io.strimzi.kafka.oauth.client"
                            ".JaasClientOauthLoginCallbackHandler",
                            "topic": (
                                self._context.get_se_streaming_stats_topic_name
                                if self._context.get_env == "local"
                                else secret_handler.get_secret(
                                    self._context.get_topic_name
                                )
                            ),
                        }
                    )
                )

                _sink_hook.writer(
                    _write_args={
                        "product_id": self._context.product_id,
                        "enable_se_streaming": _se_stats_dict[
                            user_config.se_enable_streaming
                        ],
                        "kafka_write_options": kafka_write_options,
                        "stats_df": df,
                    }
                )
            else:
                _log.info(
                    "Streaming stats to kafka is disabled, hence skipping writing to kafka"
                )

        except Exception as e:
            raise SparkExpectationsMiscException(
                f"error occurred while saving the data into the stats table {e}"
            )

    def write_error_records_final(
        self, df: DataFrame, error_table: str, rule_type: str
    ) -> Tuple[int, DataFrame]:
        try:
            _log.info("_write_error_records_final started")

            failed_records = [
                f"size({dq_column}) != 0"
                for dq_column in df.columns
                if dq_column.startswith(f"{rule_type}")
            ]

            failed_records_rules = " or ".join(failed_records)
            # df = df.filter(expr(failed_records_rules))

            df = df.withColumn(
                f"meta_{rule_type}_results",
                when(
                    expr(failed_records_rules),
                    array(
                        *[
                            _col
                            for _col in df.columns
                            if _col.startswith(f"{rule_type}")
                        ]
                    ),
                ).otherwise(array(create_map())),
            ).drop(*[_col for _col in df.columns if _col.startswith(f"{rule_type}")])

            df = (
                df.withColumn(
                    f"meta_{rule_type}_results",
                    remove_empty_maps(df[f"meta_{rule_type}_results"]),
                )
                .withColumn(
                    self._context.get_run_id_name, lit(self._context.get_run_id)
                )
                .withColumn(
                    self._context.get_run_date_name,
                    lit(self._context.get_run_date),
                )
            )
            error_df = df.filter(f"size(meta_{rule_type}_results) != 0")
            self._context.print_dataframe_with_debugger(error_df)

            self.save_df_as_table(
                error_df,
                error_table,
                self._context.get_target_and_error_table_writer_config,
            )

            _error_count = error_df.count()
            if _error_count > 0:
                self.generate_summarised_row_dq_res(error_df, rule_type)

            _log.info("_write_error_records_final ended")
            return _error_count, df

        except Exception as e:
            raise SparkExpectationsMiscException(
                f"error occurred while saving data into the final error table {e}"
            )

    def generate_summarised_row_dq_res(self, df: DataFrame, rule_type: str) -> None:
        """
        This function implements/supports summarising row dq error result
        Args:
            df: error dataframe(DataFrame)
            rule_type: type of the rule(str)

        Returns:
            None

        """
        try:
            df_exploded = df.select(
                explode(f"meta_{rule_type}_results").alias("row_dq_res")
            )

            keys = (
                df_exploded.select(explode("row_dq_res"))
                .select("key")
                .distinct()
                .rdd.flatMap(lambda x: x)
                .collect()
            )
            nested_keys = [col("row_dq_res").getItem(k).alias(k) for k in keys]

            df_select = df_exploded.select(*nested_keys)
            df_pivot = (
                df_select.groupBy(df_select.columns)
                .count()
                .withColumnRenamed("count", "failed_row_count")
            )

            keys += ["failed_row_count"]
            summarised_row_dq_list = df_pivot.rdd.map(
                lambda x: {i: x[i] for i in keys}
            ).collect()

            self._context.set_summarised_row_dq_res(summarised_row_dq_list)

        except Exception as e:
            raise SparkExpectationsMiscException(
                f"error occurred created summarised row dq statistics {e}"
            )

    def generate_rules_exceeds_threshold(self, rules: dict) -> None:
        """
        This function implements/supports summarising row dq error threshold
        Args:
            rules: accepts rule metadata within dict
        Returns:
            None
        """
        try:
            error_threshold_list = []
            rules_failed_row_count: Dict[str, int] = {}
            if self._context.get_summarised_row_dq_res is None:
                return None

            rules_failed_row_count = {
                itr["rule"]: int(itr["failed_row_count"])
                for itr in self._context.get_summarised_row_dq_res
            }

            for rule in rules[f"{self._context.get_row_dq_rule_type_name}_rules"]:
                # if (
                #         not rule["enable_error_drop_alert"]
                #         or rule["rule"] not in rules_failed_row_count.keys()
                # ):
                #     continue  # pragma: no cover

                rule_name = rule["rule"]
                rule_action = rule["action_if_failed"]
                if rule_name in rules_failed_row_count.keys():
                    failed_row_count = int(rules_failed_row_count[rule_name])
                else:
                    failed_row_count = 0

                if failed_row_count is not None and failed_row_count > 0:
                    error_drop_percentage = round(
                        (failed_row_count / self._context.get_input_count) * 100, 2
                    )
                    error_threshold_list.append(
                        {
                            "rule_name": rule_name,
                            "action_if_failed": rule_action,
                            "description": rule["description"],
                            "rule_type": rule["rule_type"],
                            "error_drop_threshold": str(rule["error_drop_threshold"]),
                            "error_drop_percentage": str(error_drop_percentage),
                        }
                    )

            if len(error_threshold_list) > 0:
                self._context.set_rules_exceeds_threshold(error_threshold_list)

        except Exception as e:
            raise SparkExpectationsMiscException(
                f"An error occurred while creating error threshold list : {e}"
            )



# pylint: disable=too-many-lines
import os
from datetime import timezone
from datetime import datetime
from dataclasses import dataclass
from uuid import uuid1
from typing import Dict, Optional, List
from pyspark.sql import DataFrame, SparkSession
from spark_expectations.config.user_config import Constants as user_config
from spark_expectations.core.exceptions import SparkExpectationsMiscException


# TODO: Add exceptions to follow standardized naming conventions for _run_id, product_id and _dq_stats_table_name in
#       the future
@dataclass
class SparkExpectationsContext:
    """
    This class provides the context for SparkExpectations
    """

    product_id: str
    spark: SparkSession

    def __post_init__(self) -> None:
        self._run_id: str = f"{self.product_id}_{uuid1()}"
        self._run_date: str = self.set_run_date()
        self._dq_stats_table_name: Optional[str] = None
        self._final_table_name: Optional[str] = None
        self._error_table_name: Optional[str] = None
        self._row_dq_rule_type_name: str = "row_dq"
        self._agg_dq_rule_type_name: str = "agg_dq"
        self._query_dq_rule_type_name: str = "query_dq"
        self._row_dq_status: str = "Skipped"
        self._source_agg_dq_status: str = "Skipped"
        self._final_agg_dq_status: str = "Skipped"
        self._source_query_dq_status: str = "Skipped"
        self._final_query_dq_status: str = "Skipped"
        self._dq_run_status: str = "Failed"

        # above configuration variable value has to be set to python
        self._dq_project_env_name = "spark_expectations"
        self._dq_config_file_name = "dq_spark_expectations_config.ini"
        self._dq_config_abs_path: Optional[str] = None

        self._enable_mail: bool = False
        self._to_mail: Optional[str] = None
        self._mail_subject: Optional[str] = None
        self._mail_from: Optional[str] = None
        self._mail_smtp_server: str
        self._mail_smtp_port: int

        self._enable_slack: bool = False
        self._slack_webhook_url: Optional[str] = None

        self._enable_teams: bool = False
        self._teams_webhook_url: Optional[str] = None

        self._table_name: Optional[str] = None
        self._input_count: int = 0
        self._error_count: int = 0
        self._output_count: int = 0

        self._env: Optional[str] = None

        self._se_streaming_stats_topic_name: str = "dq-sparkexpectations-stats"
        self._se_streaming_row_dq_res_topic_name: str = (
            "dq-sparkexpectations-row-dq-results"
        )

        self._source_agg_dq_result: Optional[List[Dict[str, str]]] = None
        self._final_agg_dq_result: Optional[List[Dict[str, str]]] = None
        self._source_query_dq_result: Optional[List[Dict[str, str]]] = None
        self._final_query_dq_result: Optional[List[Dict[str, str]]] = None

        self._notification_on_start: bool = False
        self._notification_on_completion: bool = False
        self._notification_on_fail: bool = False
        self._error_drop_threshold: int = 100

        self._cerberus_url: str = "your_cerberus_url"
        self._cerberus_cred_path: str = "your_cerberus_sdb_path"
        self._cerberus_token: Optional[str] = os.environ.get(
            "DQ_SPARK_EXPECTATIONS_CERBERUS_TOKEN"
        )

        self._se_streaming_stats_dict: Dict[str, str]
        self._enable_se_streaming: bool = False
        self._se_streaming_secret_env: str = ""

        self._debugger_mode: bool = False
        self._supported_df_query_dq: DataFrame = self.set_supported_df_query_dq()

        self._source_agg_dq_start_time: Optional[datetime] = None
        self._final_agg_dq_start_time: Optional[datetime] = None
        self._source_query_dq_start_time: Optional[datetime] = None
        self._final_query_dq_start_time: Optional[datetime] = None
        self._row_dq_start_time: Optional[datetime] = None
        self._dq_start_time: Optional[datetime] = None

        self._source_agg_dq_end_time: Optional[datetime] = None
        self._final_agg_dq_end_time: Optional[datetime] = None
        self._source_query_dq_end_time: Optional[datetime] = None
        self._final_query_dq_end_time: Optional[datetime] = None
        self._row_dq_end_time: Optional[datetime] = None
        self._dq_end_time: Optional[datetime] = None

        self._run_id_name = "meta_dq_run_id"
        self._run_date_name = "meta_dq_run_date"
        self._run_date_time_name = "meta_dq_run_datetime"

        self._num_row_dq_rules: int = 0
        self._num_agg_dq_rules: Dict[str, int] = {
            "num_agg_dq_rules": 0,
            "num_source_agg_dq_rules": 0,
            "num_final_agg_dq_rules": 0,
        }
        self._num_query_dq_rules: Dict[str, int] = {
            "num_query_dq_rules": 0,
            "num_source_query_dq_rules": 0,
            "num_final_query_dq_rules": 0,
        }
        self._num_dq_rules: int = 0
        self._summarised_row_dq_res: Optional[List[Dict[str, str]]] = None
        self._rules_error_per: Optional[List[dict]] = None

        self._target_and_error_table_writer_config: dict = {}
        self._stats_table_writer_config: dict = {}

    @property
    def get_run_id(self) -> str:
        """
        Get run_id for the instance of spark-expectations class

        Returns:
            str: returns the run_id
        """
        return self._run_id

    @property
    def get_run_date(self) -> str:
        """
        Get run_date for the instance of the spark-expectations class

        Returns:
            str: returns the run_date

        """
        return self._run_date

    def set_dq_stats_table_name(self, dq_stats_table_name: str) -> None:
        self._dq_stats_table_name = dq_stats_table_name

    @property
    def get_dq_stats_table_name(self) -> str:
        """
        Get dq_stats_table_name to which the final stats of the dq job will be written into

        Returns:
            str: returns the dq_stats_table_name
        """
        if self._dq_stats_table_name:
            return self._dq_stats_table_name
        raise SparkExpectationsMiscException(
            """The spark expectations context is not set completely, please assign '_dq_stats_table_name' before 
            accessing it"""
        )

    def set_final_table_name(self, final_table_name: str) -> None:
        self._final_table_name = final_table_name

    @property
    def get_final_table_name(self) -> str:
        """
        Get dq_stats_table_name to which the final stats of the dq job will be written into

        Returns:
            str: returns the dq_stats_table_name
        """
        if self._final_table_name:
            return self._final_table_name
        raise SparkExpectationsMiscException(
            """The spark expectations context is not set completely, please assign '_final_table_name' before 
            accessing it"""
        )

    def set_error_table_name(self, error_table_name: str) -> None:
        self._error_table_name = error_table_name

    @property
    def get_error_table_name(self) -> str:
        """
        Get dq_stats_table_name to which the final stats of the dq job will be written into

        Returns:
            str: returns the dq_stats_table_name
        """
        if self._error_table_name:
            return self._error_table_name
        raise SparkExpectationsMiscException(
            """The spark expectations context is not set completely, please assign '_error_table_name' before 
            accessing it"""
        )

    @staticmethod
    def set_run_date() -> str:
        """
        This function is used to generate the current datatime in UTC

        Returns:
            str: Returns the current utc datatime in the format - "%Y-%m-%d %H:%M:%S"

        """
        current_datetime: datetime = datetime.now(timezone.utc)
        return current_datetime.replace(tzinfo=timezone.utc).strftime(
            "%Y-%m-%d %H:%M:%S"
        )

    @property
    def get_row_dq_rule_type_name(self) -> str:
        """
        This function is used to get row data quality rule type name

        Returns:
             str: Returns _row_dq_rule_type_name"

        """
        if self._row_dq_rule_type_name:
            return self._row_dq_rule_type_name
        raise SparkExpectationsMiscException(
            """The spark expectations context is not set completely, please assign '_row_dq_rule_type_name' before 
            accessing it"""
        )

    @property
    def get_agg_dq_rule_type_name(self) -> str:
        """
        This function is used to get aggregation data quality rule type name

        Returns:
            str: Returns _agg_dq_rule_type_name"

        """
        if self._agg_dq_rule_type_name:
            return self._agg_dq_rule_type_name
        raise SparkExpectationsMiscException(
            """The spark expectations context is not set completely, please assign '_agg_dq_rule_type_name' before 
            accessing it"""
        )

    @property
    def get_query_dq_rule_type_name(self) -> str:
        """
        This function is used to get query data quality rule type name

        Returns:
            str: Returns _query_dq_rule_type_name"

        """
        if self._query_dq_rule_type_name:
            return self._query_dq_rule_type_name
        raise SparkExpectationsMiscException(
            """The spark expectations context is not set completely, please assign '_query_dq_rule_type_name' before 
            accessing it"""
        )

    def set_row_dq_status(self, row_dq_status: str = "Skipped") -> None:
        self._row_dq_status = row_dq_status

    @property
    def get_row_dq_status(self) -> str:
        """
        This function is used to get row data quality status

        Returns:
            str: Returns _row_dq_status"

        """
        if self._row_dq_status:
            return self._row_dq_status
        raise SparkExpectationsMiscException(
            """The spark expectations context is not set completely, please assign '_row_dq_status' before 
            accessing it"""
        )

    def set_source_agg_dq_status(self, source_agg_dq_status: str = "Skipped") -> None:
        self._source_agg_dq_status = source_agg_dq_status

    @property
    def get_source_agg_dq_status(self) -> str:
        """
        This function is used to get source aggregation data quality status

        Returns:
            str: Returns _source_agg_dq_status"

        """
        if self._source_agg_dq_status:
            return self._source_agg_dq_status
        raise SparkExpectationsMiscException(
            """The spark expectations context is not set completely, please assign '_source_agg_dq_status' before 
            accessing it"""
        )

    def set_final_agg_dq_status(self, final_agg_dq_status: str = "Skipped") -> None:
        self._final_agg_dq_status = final_agg_dq_status

    @property
    def get_final_agg_dq_status(self) -> str:
        """
        This function is used to get final aggregation data quality status

        Returns:
            str: Returns _final_agg_dq_status"

        """
        if self._final_agg_dq_status:
            return self._final_agg_dq_status
        raise SparkExpectationsMiscException(
            """The spark expectations context is not set completely, please assign '_final_agg_dq_status' before 
            accessing it"""
        )

    def set_source_query_dq_status(
        self, source_query_dq_status: str = "Skipped"
    ) -> None:
        self._source_query_dq_status = source_query_dq_status

    @property
    def get_source_query_dq_status(self) -> str:
        """
        This function is used to get source query data quality status

        Returns:
            str: Returns _source_query_dq_status"

        """
        if self._source_query_dq_status:
            return self._source_query_dq_status
        raise SparkExpectationsMiscException(
            """The spark expectations context is not set completely, please assign '_source_query_dq_status' before 
            accessing it"""
        )

    def set_final_query_dq_status(self, final_query_dq_status: str = "Skipped") -> None:
        self._final_query_dq_status = final_query_dq_status

    @property
    def get_final_query_dq_status(self) -> str:
        """
        This function is used to get final query dq  data quality status

        Returns:
            str: Returns _final_query_dq_status"

        """
        if self._final_query_dq_status:
            return self._final_query_dq_status
        raise SparkExpectationsMiscException(
            """The spark expectations context is not set completely, please assign '_final_query_dq_status' before 
            accessing it"""
        )

    def set_dq_run_status(self, dq_run_status: str = "Failed") -> None:
        self._dq_run_status = dq_run_status

    @property
    def get_dq_run_status(self) -> str:
        """
        This function is used to get data quality pipeline status

        Returns:
            str: Returns _dq_status"

        """
        if self._dq_run_status:
            return self._dq_run_status
        raise SparkExpectationsMiscException(
            """The spark expectations context is not set completely, please assign '_dq_run_status' before 
            accessing it"""
        )

    @property
    def get_config_file_path(self) -> str:
        """
        This function returns config file abs path
        Returns:
            str: Returns _config_file_path(str)

        """
        if self._dq_config_abs_path:
            return self._dq_config_abs_path

        raise SparkExpectationsMiscException(
            """The spark expectations context is not set completely, please assign '_dq_config_abs_path' before
            accessing it"""
        )

    def set_mail_smtp_server(self, mail_smtp_server: str) -> None:
        self._mail_smtp_server = mail_smtp_server

    @property
    def get_mail_smtp_server(self) -> str:
        """
        This functions returns smtp server host
        Returns:
            str: returns _mail_smtp_server

        """
        if self._mail_smtp_server:
            return self._mail_smtp_server

        raise SparkExpectationsMiscException(
            """The spark expectations context is not set completely, please assign '_mail_smtp_server' before 
            accessing it"""
        )

    def set_mail_smtp_port(self, mail_smtp_port: int) -> None:
        self._mail_smtp_port = mail_smtp_port

    @property
    def get_mail_smtp_port(self) -> int:
        """
        This functions returns smtp port
        Returns:
            int: returns _mail_smtp_server port

        """
        if self._mail_smtp_port:
            return self._mail_smtp_port

        raise SparkExpectationsMiscException(
            """The spark expectations context is not set completely, please assign '_mail_smtp_port' before 
            accessing it"""
        )

    def set_enable_mail(self, enable_mail: bool) -> None:
        self._enable_mail = bool(enable_mail)

    @property
    def get_enable_mail(self) -> bool:
        """
        This function return whether mail notification to enable or not
        Returns:
            str: Returns  _enable_mail(bool)

        """
        return self._enable_mail

    def set_to_mail(self, to_mail: str) -> None:
        self._to_mail = to_mail

    @property
    def get_to_mail(self) -> str:
        """
        This function returns list of mail id's
        Returns:
            str: Returns _mail_id(str)

        """
        if self._to_mail:
            return self._to_mail

        raise SparkExpectationsMiscException(
            """The spark expectations context is not set completely, please assign '_to_mail' before 
            accessing it"""
        )

    def set_mail_from(self, mail_from: str) -> None:
        self._mail_from = mail_from

    @property
    def get_mail_from(self) -> str:
        """
        This function returns mail id to send email
        Returns:

        """
        if self._mail_from:
            return self._mail_from

        raise SparkExpectationsMiscException(
            """The spark expectations context is not set completely, please assign '_mail_from' before 
            accessing it"""
        )

    def set_mail_subject(self, mail_subject: str) -> None:
        self._mail_subject = mail_subject

    @property
    def get_mail_subject(self) -> str:
        """
        This function returns mail subject
        Returns:
            str: Returns _mail_subject(str)

        """
        if self._mail_subject:
            return self._mail_subject

        raise SparkExpectationsMiscException(
            """The spark expectations context is not set completely, please assign '_mail_subject' before 
            accessing it"""
        )

    def set_enable_slack(self, enable_slack: bool) -> None:
        """

        Args:
            enable_slack:

        Returns:

        """
        self._enable_slack = enable_slack

    @property
    def get_enable_slack(self) -> bool:
        """
        This function returns whether to enable slack notification or not
        Returns: Returns _enable_slack(bool)

        """
        return self._enable_slack

    def set_slack_webhook_url(self, slack_webhook_url: str) -> None:
        self._slack_webhook_url = slack_webhook_url

    @property
    def get_slack_webhook_url(self) -> str:
        """
        This function returns sack webhook url
        Returns:
            str: Returns _webhook_url(str)

        """

        if self._slack_webhook_url:
            return self._slack_webhook_url
        raise SparkExpectationsMiscException(
            """The spark expectations context is not set completely, please assign '_slack_webhook_url' before 
            accessing it"""
        )

    def set_enable_teams(self, enable_teams: bool) -> None:
        """

        Args:
            enable_teams:

        Returns:

        """
        self._enable_teams = enable_teams

    @property
    def get_enable_teams(self) -> bool:
        """
        This function returns whether to enable teams notification or not
        Returns: Returns _enable_teams(bool)

        """
        return self._enable_teams

    def set_teams_webhook_url(self, teams_webhook_url: str) -> None:
        self._teams_webhook_url = teams_webhook_url

    @property
    def get_teams_webhook_url(self) -> str:
        """
        This function returns sack webhook url
        Returns:
            str: Returns _webhook_url(str)

        """

        if self._teams_webhook_url:
            return self._teams_webhook_url
        raise SparkExpectationsMiscException(
            """The spark expectations context is not set completely, please assign '_teams_webhook_url' before 
            accessing it"""
        )

    def set_table_name(self, table_name: str) -> None:
        self._table_name = table_name

    @property
    def get_table_name(self) -> str:
        """
        This function returns table name
        Returns:
            str: Returns _table_name(str)

        """
        if self._table_name:
            return self._table_name
        raise SparkExpectationsMiscException(
            """The spark expectations context is not set completely, please assign '_table_name' before 
            accessing it"""
        )

    def set_input_count(self, input_count: int = 0) -> None:
        self._input_count = input_count

    @property
    def get_input_count(self) -> int:
        """
        This function return input count
        Returns:
            int: Returns _input_count(int)

        """
        return self._input_count

    def set_error_count(self, error_count: int = 0) -> None:
        self._error_count = error_count

    @property
    def get_error_count(self) -> int:
        """
        This functions return error count
        Returns:
            int: Returns _error_count(int)

        """
        return self._error_count

    def set_output_count(self, output_count: int = 0) -> None:
        self._output_count = output_count

    @property
    def get_output_count(self) -> int:
        """
        This function returns output count
        Returns:
            int: Returns _output(int)

        """
        return self._output_count

    def set_source_agg_dq_result(
        self, source_agg_dq_result: Optional[List[Dict[str, str]]] = None
    ) -> None:
        self._source_agg_dq_result = source_agg_dq_result

    @property
    def get_source_agg_dq_result(self) -> Optional[List[Dict[str, str]]]:
        """
        This function return status of the source_agg_dq_result
        Returns:
            dict: Returns source_agg_dq_result which in list of dict with str(key) and str(value)

        """

        return self._source_agg_dq_result

    def set_final_agg_dq_result(
        self, final_agg_dq_result: Optional[List[Dict[str, str]]] = None
    ) -> None:
        self._final_agg_dq_result = final_agg_dq_result

    @property
    def get_final_agg_dq_result(self) -> Optional[List[Dict[str, str]]]:
        """
        This function return status of the final_agg_dq_result
        Returns:
            dict: Returns final_agg_dq_result which in list of dict with str(key) and str(value)

        """
        return self._final_agg_dq_result

    def set_source_query_dq_result(
        self, source_query_dq_result: Optional[List[Dict[str, str]]] = None
    ) -> None:
        self._source_query_dq_result = source_query_dq_result

    @property
    def get_source_query_dq_result(self) -> Optional[List[Dict[str, str]]]:
        """
        This function return status of the source_query_dq_result
        Returns:
            dict: Returns source_query_dq_result which in list of dict with str(key) and str(value)

        """
        return self._source_query_dq_result

    def set_final_query_dq_result(
        self, final_query_dq_result: Optional[List[Dict[str, str]]] = None
    ) -> None:
        self._final_query_dq_result = final_query_dq_result

    @property
    def get_final_query_dq_result(self) -> Optional[List[Dict[str, str]]]:
        """
        This function return status of the final_query_dq_result
        Returns:
            dict: Returns final_query_dq_result which in list of dict with str(key) and str(value)

        """
        return self._final_query_dq_result

    def set_notification_on_start(self, notification_on_start: bool) -> None:
        self._notification_on_start = notification_on_start

    @property
    def get_notification_on_start(self) -> bool:
        """
        This function returns notification on start
        Returns:
            bool: Returns _notification_on_start

        """
        return self._notification_on_start

    def set_notification_on_completion(self, notification_on_completion: bool) -> None:
        self._notification_on_completion = notification_on_completion

    @property
    def get_notification_on_completion(self) -> bool:
        """
        This function returns notification on completion
        Returns:
            bool: Returns _notification_on_completion

        """
        return self._notification_on_completion

    def set_notification_on_fail(self, notification_on_fail: bool) -> None:
        self._notification_on_fail = notification_on_fail

    @property
    def get_notification_on_fail(self) -> bool:
        """
        This function returns notification on fail
        Returns:
            bool: Returns _notification_on_fail

        """
        return self._notification_on_fail

    def set_env(self, env: Optional[str]) -> None:
        """
        Args:
            env: which accepts env type

        Returns:
            None

        """
        self._env = env

    @property
    def get_env(self) -> Optional[str]:
        """
        functions returns running environment type
        Returns:
               str: Returns _env
        """
        return self._env

    @property
    def get_error_percentage(self) -> float:
        """
        This function returns error percentage
        Returns:
               float: error percentage
        """
        if self._input_count > 0:
            return round((self.get_error_count / self.get_input_count) * 100, 2)
        return 0.0

    @property
    def get_output_percentage(self) -> float:
        """
        This function return output percentage
        Returns:
            float: output percentage

        """
        if self._input_count > 0:
            return round((self.get_output_count / self.get_input_count) * 100, 2)
        return 0.0

    @property
    def get_success_percentage(self) -> float:
        """
        This function returns success percentage
        Returns:
            float: success percentage

        """
        if self._input_count > 0:
            return round(
                ((self.get_input_count - self.get_error_count) / self.get_input_count)
                * 100,
                2,
            )
        return 0.0

    @property
    def get_error_drop_percentage(self) -> float:
        """
        This function returns error drop percentage percentage
        Returns:
            float: error drop percentage

        """
        if self._input_count > 0:
            return round(
                ((self.get_input_count - self.get_output_count) / self.get_input_count)
                * 100,
                2,
            )
        return 0.0

    def set_error_drop_threshold(self, error_drop_threshold: int) -> None:
        self._error_drop_threshold = error_drop_threshold

    @property
    def get_error_drop_threshold(self) -> int:
        """
        This function return error threshold breach
        Returns:
               int: error threshold breach
        """
        if self._error_drop_threshold:
            return self._error_drop_threshold
        raise SparkExpectationsMiscException(
            """The spark expectations context is not set completely, please assign '_error_drop_threshold'  before 
            accessing it"""
        )

    @property
    def get_cerberus_url(self) -> str:
        """
        This functions implemented to return cerberus url
        Returns:

        """
        if self._cerberus_url:
            return self._cerberus_url
        raise SparkExpectationsMiscException(
            """The spark expectations context is not set completely, please assign '_cerberus_url'  before 
            accessing it"""
        )

    @property
    def get_cerberus_cred_path(self) -> str:
        """
        This functions implemented to return cerberus credentials path
        Returns:

        """
        if self._cerberus_cred_path:
            return self._cerberus_cred_path
        raise SparkExpectationsMiscException(
            """The spark expectations context is not set completely, please assign '_cerberus_cred_path'  before 
            accessing it"""
        )

    @property
    def get_cerberus_token(self) -> str:
        """
        This functions implemented to return cerberus token
        Returns:

        """
        if self._cerberus_token:
            return self._cerberus_token
        raise SparkExpectationsMiscException(
            """The spark expectations context is not set completely, please assign '_cerberus_token'  before 
            accessing it"""
        )

    def set_se_streaming_stats_dict(
        self, se_streaming_stats_dict: Dict[str, str]
    ) -> None:
        """
        This function helps to set secret keys dict"""
        self._se_streaming_stats_dict = se_streaming_stats_dict

    @property
    def get_se_streaming_stats_dict(self) -> Dict[str, str]:
        """
        This function returns secret keys dict
        """
        return self._se_streaming_stats_dict

    @property
    def get_secret_type(self) -> Optional[str]:
        """
        This function helps in getting secret type
        Returns:
             secret type in Optional[str]

        """
        if self._se_streaming_stats_dict.get(user_config.secret_type):
            return self._se_streaming_stats_dict[user_config.secret_type].lower()
        raise SparkExpectationsMiscException(
            """The spark expectations context is not set completely, please assign 
            'UserConfig.secret_type' before 
            accessing it"""
        )

    @property
    def get_server_url_key(self) -> Optional[str]:
        """
        This function helps in getting key / path for kafka server url
         Returns:
              kafka server url key / path in Optional[str]
        """
        _server_url_key: Optional[str] = (
            self._se_streaming_stats_dict.get(user_config.cbs_kafka_server_url)
            if self.get_secret_type == "cerberus"
            else self._se_streaming_stats_dict.get(user_config.dbx_kafka_server_url)
        )
        if _server_url_key:
            return _server_url_key
        raise SparkExpectationsMiscException(
            """The spark expectations context is not set completely, please assign
            'UserConfig.cbs_kafka_server_url' before
            accessing it"""
        )

    @property
    def get_token_endpoint_url(self) -> Optional[str]:
        """
        This function helps in getting key / path for end point url
         Returns:
             end point url key / path in  Optional[str]
        """
        _token_endpoint_url: Optional[str] = (
            self._se_streaming_stats_dict.get(user_config.cbs_secret_token_url)
            if self.get_secret_type == "cerberus"
            else self._se_streaming_stats_dict.get(user_config.dbx_secret_token_url)
        )
        if _token_endpoint_url:
            return _token_endpoint_url
        raise SparkExpectationsMiscException(
            """The spark expectations context is not set completely, please assign
            'UserConfig.cbs_secret_token_url' before
            accessing it"""
        )

    @property
    def get_token(self) -> Optional[str]:
        """
        This function helps in getting key / path for token
        Returns:
             token key / path in Optional[str]
        """
        _token: Optional[str] = (
            self._se_streaming_stats_dict.get(user_config.cbs_secret_token)
            if self.get_secret_type == "cerberus"
            else self._se_streaming_stats_dict.get(user_config.dbx_secret_token)
        )
        if _token:
            return _token
        raise SparkExpectationsMiscException(
            """The spark expectations context is not set completely, please assign
            'UserConfig.cbs_secret_token' before
            accessing it"""
        )

    @property
    def get_client_id(self) -> Optional[str]:
        """
        This function helps in getting key / path for client id
        Returns:
            client id key / path in Optional[str]
        """
        _client_id: Optional[str] = (
            self._se_streaming_stats_dict.get(user_config.cbs_secret_app_name)
            if self.get_secret_type == "cerberus"
            else self._se_streaming_stats_dict.get(user_config.dbx_secret_app_name)
        )
        if _client_id:
            return _client_id
        raise SparkExpectationsMiscException(
            """The spark expectations context is not set completely, please assign
            'UserConfig.cbs_secret_app_name' before
            accessing it"""
        )

    @property
    def get_topic_name(self) -> Optional[str]:
        """
        This function helps in getting key / path for topic name
        Returns:
            topic name key / path in Optional[str]
        """
        _topic_name: Optional[str] = (
            self._se_streaming_stats_dict.get(user_config.cbs_topic_name)
            if self.get_secret_type == "cerberus"
            else self._se_streaming_stats_dict.get(user_config.dbx_topic_name)
        )
        if _topic_name:
            return _topic_name
        raise SparkExpectationsMiscException(
            """The spark expectations context is not set completely, please assign 
            'UserConfig.cbs_topic_name' before 
            accessing it"""
        )

    def set_se_streaming_stats_topic_name(
        self, se_streaming_stats_topic_name: str
    ) -> None:
        self._se_streaming_stats_topic_name = se_streaming_stats_topic_name

    @property
    def get_se_streaming_stats_topic_name(self) -> str:
        """
        This function returns kafka topic name
        Returns:
            str: Returns _se_streaming_stats_topic_name

        """
        if self._se_streaming_stats_topic_name:
            return self._se_streaming_stats_topic_name
        raise SparkExpectationsMiscException(
            """The spark expectations context is not set completely, please assign 
            '_se_streaming_stats_topic_name' before 
            accessing it"""
        )

    def set_debugger_mode(self, debugger_mode: bool) -> None:
        """
        This function sets debugger mode
        Returns:

        """
        self._debugger_mode = debugger_mode

    @property
    def get_debugger_mode(self) -> bool:
        """
        This function returns a debugger
        Returns:
             bool: return debugger

        """
        return self._debugger_mode

    def print_dataframe_with_debugger(self, df: DataFrame) -> None:
        """
        This function has a debugger that can print out the DataFrame
        Returns:

        """
        if self.get_debugger_mode:
            df.show(truncate=False)

    def set_supported_df_query_dq(self) -> DataFrame:
        return self.spark.createDataFrame(
            [
                {
                    "spark_expectations_query_check": "supported_place_holder_dataset_to_run_query_check"
                }
            ]
        )

    @property
    def get_supported_df_query_dq(self) -> DataFrame:
        """
        This function returns the place holder dataframe for query check
        Returns:
            DataFrame: returns dataframe for query dq

        """

        if self._supported_df_query_dq:
            return self._supported_df_query_dq
        raise SparkExpectationsMiscException(
            """The spark expectations context is not set completely, please assign '_supported_df_query_dq'  before 
            accessing it"""
        )

    def set_source_agg_dq_start_time(self) -> None:
        """
        This function sets start time source agg dq computation
        Returns:
             None

        """
        self._source_agg_dq_start_time = datetime.now()

    def set_source_agg_dq_end_time(self) -> None:
        """
        This function sets end time source agg dq computation
        Returns:
            None

        """
        self._source_agg_dq_end_time = datetime.now()

    def set_final_agg_dq_start_time(self) -> None:
        """
        This function sets start time final agg dq computation
        Returns:
        None

        """
        self._final_agg_dq_start_time = datetime.now()

    def set_final_agg_dq_end_time(self) -> None:
        """
        This function sets end time final agg dq computation
        Returns:
            None
        """
        self._final_agg_dq_end_time = datetime.now()

    def set_source_query_dq_start_time(self) -> None:
        """
        This function sets start time source query dq computation
        Returns:
            None
        """
        self._source_query_dq_start_time = datetime.now()

    def set_source_query_dq_end_time(self) -> None:
        """
        This function sets end time source query dq computation
        Returns:
            None
        """
        self._source_query_dq_end_time = datetime.now()

    def set_final_query_dq_start_time(self) -> None:
        """
        This function sets start time final query dq computation
        Returns:
            None
        """
        self._final_query_dq_start_time = datetime.now()

    def set_final_query_dq_end_time(self) -> None:
        """
        This function sets end time final query dq computation
        Returns:
            None
        """
        self._final_query_dq_end_time = datetime.now()

    def set_row_dq_start_time(self) -> None:
        """
        This function sets start time row dq computation
        Returns:
            None
        """
        self._row_dq_start_time = datetime.now()

    def set_row_dq_end_time(self) -> None:
        """
        This function sets end time row dq computation
        Returns:
            None
        """
        self._row_dq_end_time = datetime.now()

    def set_dq_start_time(self) -> None:
        """
        This function sets start time dq computation
        Returns:
            None
        """
        self._dq_start_time = datetime.now()

    def set_dq_end_time(self) -> None:
        """
        This function sets end time dq computation
        Returns:
            None
        """
        self._dq_end_time = datetime.now()

    def set_end_time_when_dq_job_fails(self) -> None:
        """
        function used to set end time when job fails in any one of the stages by using start time
        Returns:
        """
        if self._source_agg_dq_start_time and self._source_agg_dq_end_time is None:
            self.set_source_agg_dq_end_time()
        elif (
            self._source_query_dq_start_time and self._source_query_dq_end_time is None
        ):
            self.set_source_query_dq_end_time()
        elif self._row_dq_start_time and self._row_dq_end_time is None:
            self.set_row_dq_end_time()
        elif self._final_agg_dq_start_time and self._final_agg_dq_end_time is None:
            self.set_final_agg_dq_end_time()
        elif self._final_query_dq_start_time and self._final_query_dq_end_time is None:
            self.set_final_query_dq_end_time()

    def get_time_diff(
        self, start_time: Optional[datetime], end_time: Optional[datetime]
    ) -> float:
        """
        This function implements time diff
        Args:
            start_time:
            end_time:

        Returns:

        """
        if start_time and end_time:
            time_diff = end_time - start_time

            return round(float(time_diff.total_seconds()), 1)
        else:
            return 0.0

    @property
    def get_source_agg_dq_run_time(self) -> float:
        """
        This function implements time diff for source agg dq run
        Returns:
             float: time in float

        """
        return self.get_time_diff(
            self._source_agg_dq_start_time, self._source_agg_dq_end_time
        )

    @property
    def get_final_agg_dq_run_time(self) -> float:
        """
        This function implements time diff for final agg dq run
        Returns:
             float: time in float

        """
        return self.get_time_diff(
            self._final_agg_dq_start_time, self._final_agg_dq_end_time
        )

    @property
    def get_source_query_dq_run_time(self) -> float:
        """
        This function implements time diff for source query dq run
        Returns:
            float: time in float
        """
        return self.get_time_diff(
            self._source_query_dq_start_time, self._source_query_dq_end_time
        )

    @property
    def get_final_query_dq_run_time(self) -> float:
        """
        This function implements time diff for final query dq run
        Returns:
            float: time in float
        """
        return self.get_time_diff(
            self._final_query_dq_start_time, self._final_query_dq_end_time
        )

    @property
    def get_row_dq_run_time(self) -> float:
        """
        This function implements time diff for row dq run
        Returns:
            float: time in float
        """
        return self.get_time_diff(self._row_dq_start_time, self._row_dq_end_time)

    @property
    def get_dq_run_time(self) -> float:
        """
        This function implements time diff for dq run
        Returns:
            float: time in float
        """
        return self.get_time_diff(self._dq_start_time, self._dq_end_time)

    @property
    def get_run_id_name(self) -> str:
        """
        This function returns name for the run_id column
        Returns:
            str: name of run_id in str

        """
        if self._run_id_name:
            return self._run_id_name
        raise SparkExpectationsMiscException(
            """The spark expectations context is not set completely, please assign '_run_id_name'  before 
            accessing it"""
        )

    @property
    def get_run_date_name(self) -> str:
        """
        This function returns name for the run_date column
        Returns:
            str: name of run_date in str

        """
        if self._run_date_name:
            return self._run_date_name
        raise SparkExpectationsMiscException(
            """The spark expectations context is not set completely, please assign '_run_date_name'  before 
            accessing it"""
        )

    @property
    def get_run_date_time_name(self) -> str:
        """
        This function returns name for the run_date_time column
        Returns:
            str: name of run_date_time in str

        """
        if self._run_date_time_name:
            return self._run_date_time_name
        raise SparkExpectationsMiscException(
            """The spark expectations context is not set completely, please assign '_run_date_time_name'  before 
            accessing it"""
        )

    def reset_num_row_dq_rules(self) -> None:
        """
        This function used to reset the _num_row_dq_rules
        Returns:
            None

        """

        self._num_row_dq_rules = 0  # pragma: no cover

    def reset_num_agg_dq_rules(self) -> None:
        """
        This function used to reset the_num_agg_dq_rules
        Returns:
            None

        """
        self._num_agg_dq_rules = {
            "num_agg_dq_rules": 0,
            "num_source_agg_dq_rules": 0,
            "num_final_agg_dq_rules": 0,
        }

    def reset_num_query_dq_rules(self) -> None:
        """
        This function used to rest the _num_query_dq_rules
        Returns:
            None

        """
        self._num_query_dq_rules = {
            "num_query_dq_rules": 0,
            "num_source_query_dq_rules": 0,
            "num_final_query_dq_rules": 0,
        }

    def reset_num_dq_rules(self) -> None:
        """
        This function used to reset the _num_dq_rules
        Returns:
            None

        """
        self._num_dq_rules = 0

    def set_num_row_dq_rules(self) -> None:
        """
        This function sets number of applied row dq rules for batch run
        Returns:
            None

        """
        self._num_row_dq_rules += 1
        self._num_dq_rules += 1

    def set_num_agg_dq_rules(
        self, source_agg_enabled: bool = False, final_agg_enabled: bool = False
    ) -> None:
        """
        This function sets number of applied agg dq rules for batch run
        source_agg_enabled: Marked True when agg rules set for source, by default False
        final_agg_enabled: Marked True when agg rules set for final, by default False
        Returns:
            None
        """

        self._num_agg_dq_rules["num_agg_dq_rules"] += 1
        self._num_dq_rules += 1

        if source_agg_enabled:
            self._num_agg_dq_rules["num_source_agg_dq_rules"] += 1
        if final_agg_enabled:
            self._num_agg_dq_rules["num_final_agg_dq_rules"] += 1

    def set_num_query_dq_rules(
        self, source_query_enabled: bool = False, final_query_enabled: bool = False
    ) -> None:
        """
        This function sets number of applied query dq rules for batch run
        source_query_enabled: Marked True when query rules set for source, by default False
        final_query_enabled: Marked True when query rules set for final, by default False
        Returns:
            None
        """

        self._num_query_dq_rules["num_query_dq_rules"] += 1
        self._num_dq_rules += 1

        if source_query_enabled:
            self._num_query_dq_rules["num_source_query_dq_rules"] += 1
        if final_query_enabled:
            self._num_query_dq_rules["num_final_query_dq_rules"] += 1

    @property
    def get_num_row_dq_rules(self) -> int:
        """
        This function returns number row dq rules applied for batch run
        Returns:
            int: number of rules in int

        """

        if isinstance(self._num_row_dq_rules, int):
            return self._num_row_dq_rules
        raise SparkExpectationsMiscException(
            """The spark expectations context is not set completely, please assign '_num_row_dq_rules'  before 
            accessing it"""
        )

    @property
    def get_num_agg_dq_rules(self) -> dict:
        """
        This function returns number agg dq rules applied for batch run
        Returns:
            int: number of rules in int

        """
        if isinstance(self._num_agg_dq_rules, dict):
            return self._num_agg_dq_rules
        raise SparkExpectationsMiscException(
            """The spark expectations context is not set completely, please assign '_num_agg_dq_rules'  before 
            accessing it"""
        )

    @property
    def get_num_query_dq_rules(self) -> dict:
        """
        This function returns number query dq rules applied for batch run
        Returns:
            int: number of rules in int

        """
        if isinstance(self._num_query_dq_rules, dict):
            return self._num_query_dq_rules
        raise SparkExpectationsMiscException(
            """The spark expectations context is not set completely, please assign '_num_query_dq_rules'  before 
            accessing it"""
        )

    @property
    def get_num_dq_rules(self) -> int:
        """
        This function returns number dq rules applied for batch run
        Returns:
            int: number of rules in int
        """
        if isinstance(self._num_dq_rules, int):
            return self._num_dq_rules
        raise SparkExpectationsMiscException(
            """The spark expectations context is not set completely, please assign '_num_dq_rules'  before 
            accessing it"""
        )

    def set_summarised_row_dq_res(
        self, summarised_row_dq_res: Optional[List[Dict[str, str]]] = None
    ) -> None:
        """
        This function implements or supports to set row dq summarised res
        Args:
            summarised_row_dq_res: list(dict)
        Returns: None

        """
        self._summarised_row_dq_res = summarised_row_dq_res

    @property
    def get_summarised_row_dq_res(self) -> Optional[List[Dict[str, str]]]:
        """
        This function returns row dq summarised res
        Returns:
            list(dict): Returns summarised_row_dq_res which in list of dict with str(key) and
            str(value) of rule meta data

        """
        return self._summarised_row_dq_res

    def set_rules_exceeds_threshold(self, rules: Optional[List[dict]] = None) -> None:
        """
        This function implements error percentage for each rule type
        """
        self._rules_error_per = rules

    @property
    def get_rules_exceeds_threshold(self) -> Optional[List[dict]]:
        """
        This function returns error percentage for each rule
        """
        return self._rules_error_per

    def set_target_and_error_table_writer_config(self, config: dict) -> None:
        """
        This function sets target and error table writer config
        Args:
            config: dict
        Returns: None

        """
        self._target_and_error_table_writer_config = config

    @property
    def get_target_and_error_table_writer_config(self) -> dict:
        """
        This function returns target and error table writer config
        Returns:
            dict: Returns target_and_error_table_writer_config which in dict

        """
        return self._target_and_error_table_writer_config

    def set_stats_table_writer_config(self, config: dict) -> None:
        """
        This function sets stats table writer config
        Args:
            config: dict
        Returns: None
        """
        self._stats_table_writer_config = config

    @property
    def get_stats_table_writer_config(self) -> dict:
        """
        This function returns stats table writer config
        Returns:
            dict: Returns stats_table_writer_config which in dict
        """
        return self._stats_table_writer_config



# mypy: ignore-errors
import os

from pyspark.sql import DataFrame
from spark_expectations import _log
from spark_expectations.examples.base_setup import set_up_delta
from spark_expectations.core.expectations import (
    SparkExpectations,
    WrappedDataFrameWriter,
)
from spark_expectations.config.user_config import Constants as user_config


writer = WrappedDataFrameWriter().mode("append").format("delta")

spark = set_up_delta()

se: SparkExpectations = SparkExpectations(
    product_id="your_product",
    rules_df=spark.table("dq_spark_local.dq_rules"),
    stats_table="dq_spark_local.dq_stats",
    stats_table_writer=writer,
    target_and_error_table_writer=writer,
    debugger=False,
    # stats_streaming_options={user_config.se_enable_streaming: False},
)


user_conf = {
    user_config.se_notifications_enable_email: False,
    user_config.se_notifications_email_smtp_host: "mailhost.com",
    user_config.se_notifications_email_smtp_port: 25,
    user_config.se_notifications_email_from: "",
    user_config.se_notifications_email_to_other_mail_id: "",
    user_config.se_notifications_email_subject: "spark expectations - data quality - notifications",
    user_config.se_notifications_enable_slack: False,
    user_config.se_notifications_slack_webhook_url: "",
    user_config.se_notifications_on_start: True,
    user_config.se_notifications_on_completion: True,
    user_config.se_notifications_on_fail: True,
    user_config.se_notifications_on_error_drop_exceeds_threshold_breach: True,
    user_config.se_notifications_on_error_drop_threshold: 15,
}


@se.with_expectations(
    target_table="dq_spark_local.customer_order",
    write_to_table=True,
    user_conf=user_conf,
    target_table_view="order",
)
def build_new() -> DataFrame:
    _df_order: DataFrame = (
        spark.read.option("header", "true")
        .option("inferSchema", "true")
        .csv(os.path.join(os.path.dirname(__file__), "resources/order.csv"))
    )
    _df_order.createOrReplaceTempView("order")

    _df_product: DataFrame = (
        spark.read.option("header", "true")
        .option("inferSchema", "true")
        .csv(os.path.join(os.path.dirname(__file__), "resources/product.csv"))
    )
    _df_product.createOrReplaceTempView("product")

    _df_customer: DataFrame = (
        spark.read.option("header", "true")
        .option("inferSchema", "true")
        .csv(os.path.join(os.path.dirname(__file__), "resources/customer.csv"))
    )

    _df_customer.createOrReplaceTempView("customer")

    return _df_order


if __name__ == "__main__":
    build_new()

    spark.sql("use dq_spark_local")
    spark.sql("select * from dq_spark_local.dq_stats").show(truncate=False)
    spark.sql("select * from dq_spark_local.dq_stats").printSchema()
    spark.sql("select * from dq_spark_local.customer_order").show(truncate=False)
    spark.sql("select count(*) from dq_spark_local.customer_order_error").show(
        truncate=False
    )

    _log.info("stats data in the kafka topic")
    # display posted statistics from the kafka topic
    spark.read.format("kafka").option(
        "kafka.bootstrap.servers", "localhost:9092"
    ).option("subscribe", "dq-sparkexpectations-stats").option(
        "startingOffsets", "earliest"
    ).option(
        "endingOffsets", "latest"
    ).load().selectExpr(
        "cast(value as string) as stats_records"
    ).show(
        truncate=False
    )

    # remove docker container
    current_dir = os.path.dirname(os.path.abspath(__file__))
    os.system(f"sh {current_dir}/docker_scripts/docker_kafka_stop_script.sh")



import os
from typing import Optional, Union, Dict
from dataclasses import dataclass

from pyspark.sql import DataFrame
from spark_expectations.core.context import SparkExpectationsContext
from spark_expectations.config.user_config import Constants as user_config
from spark_expectations.core.exceptions import (
    SparkExpectationsMiscException,
)


@dataclass
class SparkExpectationsReader:
    """
    This class implements/supports reading data from source system
    """

    _context: SparkExpectationsContext

    def __post_init__(self) -> None:
        self.spark = self._context.spark

    def set_notification_param(
        self, notification: Optional[Dict[str, Union[int, str, bool]]] = None
    ) -> None:
        """
        This function supports to read notifications configurations
        Returns: None

        """
        try:
            _default_spark_conf: Dict[str, Union[str, int, bool]] = {
                user_config.se_notifications_enable_email: False,
                user_config.se_notifications_email_smtp_host: "",
                user_config.se_notifications_email_smtp_port: 25,
                user_config.se_notifications_email_from: "",
                user_config.se_notifications_email_to_other_mail_id: "",
                user_config.se_notifications_email_subject: "spark-expectations-testing",
                user_config.se_notifications_enable_slack: False,
                user_config.se_notifications_slack_webhook_url: "",
                user_config.se_notifications_enable_teams: False,
                user_config.se_notifications_teams_webhook_url: "",
            }

            _notification_dict: Dict[str, Union[str, int, bool]] = (
                {**_default_spark_conf, **notification}
                if notification
                else _default_spark_conf
            )

            if (
                _notification_dict.get(user_config.se_notifications_enable_email)
                is True
            ):
                if (
                    _notification_dict[user_config.se_notifications_email_smtp_host]
                    and _notification_dict[user_config.se_notifications_email_from]
                    and _notification_dict[
                        user_config.se_notifications_email_to_other_mail_id
                    ]
                    and _notification_dict[user_config.se_notifications_email_subject]
                ):
                    self._context.set_enable_mail(True)
                    self._context.set_to_mail(
                        str(
                            _notification_dict[
                                user_config.se_notifications_email_to_other_mail_id
                            ]
                        )
                    )
                    self._context.set_mail_subject(
                        str(
                            _notification_dict[
                                user_config.se_notifications_email_subject
                            ]
                        )
                    )
                    self._context.set_mail_smtp_server(
                        str(
                            _notification_dict[
                                user_config.se_notifications_email_smtp_host
                            ]
                        )
                    )
                    self._context.set_mail_smtp_port(
                        int(
                            _notification_dict[
                                user_config.se_notifications_email_smtp_port
                            ]
                        )
                    )

                    self._context.set_mail_from(
                        str(_notification_dict[user_config.se_notifications_email_from])
                    )
                else:
                    raise SparkExpectationsMiscException(
                        "All params/variables required for email notification is not configured or supplied"
                    )

            if _notification_dict[user_config.se_notifications_enable_slack] is True:
                if _notification_dict[user_config.se_notifications_slack_webhook_url]:
                    self._context.set_enable_slack(True)
                    self._context.set_slack_webhook_url(
                        str(
                            _notification_dict[
                                user_config.se_notifications_slack_webhook_url
                            ]
                        )
                    )
                else:
                    raise SparkExpectationsMiscException(
                        "All params/variables required for slack notification is not configured or supplied"
                    )

            if _notification_dict[user_config.se_notifications_enable_teams] is True:
                if _notification_dict[user_config.se_notifications_teams_webhook_url]:
                    self._context.set_enable_teams(True)
                    self._context.set_teams_webhook_url(
                        str(
                            _notification_dict[
                                user_config.se_notifications_teams_webhook_url
                            ]
                        )
                    )
                else:
                    raise SparkExpectationsMiscException(
                        "All params/variables required for slack notification is not configured or supplied"
                    )

        except Exception as e:
            raise SparkExpectationsMiscException(
                f"error occurred while reading notification configurations {e}"
            )

    def get_rules_from_df(
        self,
        rules_df: DataFrame,
        target_table: str,
        is_dlt: bool = False,
        tag: Optional[str] = None,
    ) -> tuple[dict, dict]:
        """
        This function fetches the data quality rules from the table and return it as a dictionary

        Args:
            rules_df: DataFrame which has your data quality rules
            target_table: Provide the full table name for which the data quality rules are being run
            is_dlt: True if this for fetching the rules for dlt job
            tag: If is_dlt is True, provide the KPI for which you are running the data quality rule

        Returns:
            tuple: returns a tuple of two dictionaries with key as 'rule_type' and 'rules_table_row' as value in
                expectations. dict, and key as 'dq_stage_setting' and 'boolean' as value in rules_execution_settings
                    dict
        """
        try:
            self._context.set_final_table_name(target_table)
            self._context.set_error_table_name(f"{target_table}_error")
            self._context.set_table_name(target_table)
            self._context.set_env(os.environ.get("SPARKEXPECTATIONS_ENV"))

            self._context.reset_num_agg_dq_rules()
            self._context.reset_num_dq_rules()
            self._context.reset_num_row_dq_rules()
            self._context.reset_num_query_dq_rules()

            _rules_df = rules_df.filter(
                (rules_df.product_id == self._context.product_id)
                & (rules_df.table_name == target_table)
                & rules_df.is_active
            )

            self._context.print_dataframe_with_debugger(_rules_df)

            _expectations: dict = {}
            _rules_execution_settings: dict = {}
            if is_dlt:
                if tag:
                    for row in _rules_df.filter(_rules_df.tag == tag).collect():
                        _expectations[row["rule"]] = row["expectation"]
                else:
                    for row in _rules_df.collect():
                        _expectations[row["rule"]] = row["expectation"]
            else:
                for row in _rules_df.collect():
                    column_map = {
                        "product_id": row["product_id"],
                        "table_name": row["table_name"],
                        "rule_type": row["rule_type"],
                        "rule": row["rule"],
                        "column_name": row["column_name"],
                        "expectation": row["expectation"],
                        "action_if_failed": row["action_if_failed"],
                        "enable_for_source_dq_validation": row[
                            "enable_for_source_dq_validation"
                        ],
                        "enable_for_target_dq_validation": row[
                            "enable_for_target_dq_validation"
                        ],
                        "tag": row["tag"],
                        "description": row["description"],
                        "enable_error_drop_alert": row["enable_error_drop_alert"],
                        "error_drop_threshold": row["error_drop_threshold"],
                    }

                    if f"{row['rule_type']}_rules" in _expectations:
                        _expectations[f"{row['rule_type']}_rules"].append(column_map)
                    else:
                        _expectations[f"{row['rule_type']}_rules"] = [column_map]

                    # count the rules enabled for the current run
                    if row["rule_type"] == self._context.get_row_dq_rule_type_name:
                        self._context.set_num_row_dq_rules()
                    elif row["rule_type"] == self._context.get_agg_dq_rule_type_name:
                        self._context.set_num_agg_dq_rules(
                            row["enable_for_source_dq_validation"],
                            row["enable_for_target_dq_validation"],
                        )
                    elif row["rule_type"] == self._context.get_query_dq_rule_type_name:
                        self._context.set_num_query_dq_rules(
                            row["enable_for_source_dq_validation"],
                            row["enable_for_target_dq_validation"],
                        )

                    _expectations["target_table_name"] = target_table
                    _rules_execution_settings = self._get_rules_execution_settings(
                        _rules_df
                    )
            return _expectations, _rules_execution_settings
        except Exception as e:
            raise SparkExpectationsMiscException(
                f"error occurred while retrieving rules list from the table {e}"
            )

    def _get_rules_execution_settings(self, rules_df: DataFrame) -> dict:
        rules_df.createOrReplaceTempView("rules_view")
        df = self.spark.sql(
            """SELECT
                MAX(CASE WHEN rule_type = 'row_dq' THEN True ELSE False END) AS row_dq,
                MAX(CASE WHEN rule_type = 'agg_dq' AND enable_for_source_dq_validation = true  
                   THEN True ELSE False END) AS source_agg_dq,
                MAX(CASE WHEN rule_type = 'query_dq' AND enable_for_source_dq_validation = true 
                    THEN True ELSE False END) AS source_query_dq,
                MAX(CASE WHEN rule_type = 'agg_dq' AND enable_for_target_dq_validation = true 
                    THEN True ELSE False END) AS target_agg_dq,
                MAX(CASE WHEN rule_type = 'query_dq' AND enable_for_target_dq_validation = true 
                    THEN True ELSE False END) AS target_query_dq
            FROM rules_view"""
        )
        # convert the df to python dictionary as it has only one row
        rule_execution_settings = df.collect()[0].asDict()
        self.spark.catalog.dropTempView("rules_view")
        return rule_execution_settings



from typing import Dict, List, Any, Optional
from pyspark.sql import DataFrame

# from pyspark.sql.types import (
#     StringType,
#     ArrayType,
#     MapType,
# )
from pyspark.sql.functions import (
    create_map,
    expr,
    when,
    array,
    col,
    lit,
    struct,
    map_from_entries,
    array_contains,
)
from spark_expectations.utils.udf import remove_empty_maps, get_actions_list
from spark_expectations.core.context import SparkExpectationsContext
from spark_expectations.core.exceptions import (
    SparkExpectationsMiscException,
    SparkExpectOrFailException,
)


class SparkExpectationsActions:
    """
    This class implements/supports applying data quality rules on given dataframe and performing required action
    """

    @staticmethod
    def get_rule_is_active(
        _context: SparkExpectationsContext,
        rule: dict,
        _rule_type_name: str,
        _source_dq_enabled: bool = False,
        _target_dq_enabled: bool = False,
    ) -> bool:
        """
        Args:
            _context: SparkExpectationsContext class object
            rule: dict with rule properties
            _rule_type_name: which determines the type of the rule
            _source_dq_enabled: Mark it as True when dq running for source dataframe
            _target_dq_enabled: Mark it as True when dq running for target dataframe

        Returns:

        """
        _is_active: bool = False
        if (
            _rule_type_name
            in [
                _context.get_query_dq_rule_type_name,
                _context.get_agg_dq_rule_type_name,
            ]
            and _source_dq_enabled is True
        ):
            _is_active = rule["enable_for_source_dq_validation"]
        elif (
            _rule_type_name
            in [
                _context.get_query_dq_rule_type_name,
                _context.get_agg_dq_rule_type_name,
            ]
            and _target_dq_enabled is True
        ):
            _is_active = rule["enable_for_target_dq_validation"]

        return _is_active

    @staticmethod
    def create_rules_map(_rule_map: Dict[str, str]) -> Any:
        """
        This function helps to extract the selected rules properties and returns array of dict with key and value

        Args:
            _rule_map: dict with rules properties
        Returns: Array of tuple with expectations rule properties

        """
        return array(
            [
                struct(lit(elem), lit(_rule_map.get(elem)))
                for elem in [
                    "rule_type",
                    "rule",
                    "action_if_failed",
                    "tag",
                    "description",
                ]
            ]
        )

    @staticmethod
    def create_agg_dq_results(
        _context: SparkExpectationsContext, _df: DataFrame, _rule_type_name: str
    ) -> Optional[List[Dict[str, str]]]:
        """
        This function helps to collect the aggregation results in to the list
         Args:
             _context: SparkContext object
             _df: dataframe which contains agg data quality results
             _rule_type_name: which determines the type of the rule

         Returns:
                List with dict of agg rules property name and value


        """
        try:
            first_row = _df.first()
            if (
                first_row is not None
                and f"meta_{_rule_type_name}_results" in _df.columns
            ):
                meta_results = first_row[f"meta_{_rule_type_name}_results"]
                if meta_results is not None and len(meta_results) > 0:
                    return meta_results
            return None
        except Exception as e:
            raise SparkExpectationsMiscException(
                f"error occurred while running create agg dq results {e}"
            )

    @staticmethod
    def run_dq_rules(
        _context: SparkExpectationsContext,
        df: DataFrame,
        expectations: Dict[str, List[dict]],
        rule_type: str,
        _source_dq_enabled: bool = False,
        _target_dq_enabled: bool = False,
    ) -> DataFrame:
        """
        This function builds the expressions for the data quality rules and returns the dataframe with results

        Args:
            _context: Provide SparkExpectationsContext
            df: Input dataframe on which data quality rules need to be applied
            expectations: Provide the dict which has all the rules
            rule_type: identifier for the type of rule to be applied in processing
            _source_dq_enabled: Mark it as True when dq running for source dataframe
            _target_dq_enabled: Mark it as True when dq running for target dataframe

        Returns:
            DataFrame: Returns a dataframe with all the rules run the input dataframe

        """
        try:
            condition_expressions = []
            if len(expectations) <= 0:
                raise SparkExpectationsMiscException("no rules found to process")

            if (
                f"{rule_type}_rules" not in expectations
                or len(expectations[f"{rule_type}_rules"]) <= 0
            ):
                raise SparkExpectationsMiscException(
                    f"zero expectations to process for {rule_type}_rules from the `dq_rules` table, "
                    f"please configure rules or avoid this error by setting {rule_type} to False"
                )

            for rule in expectations[f"{rule_type}_rules"]:
                _rule_is_active = SparkExpectationsActions.get_rule_is_active(
                    _context,
                    rule,
                    rule_type,
                    _source_dq_enabled=_source_dq_enabled,
                    _target_dq_enabled=_target_dq_enabled,
                )

                if _rule_is_active or rule_type == _context.get_row_dq_rule_type_name:
                    column = f"{rule_type}_{rule['rule']}"
                    condition_expressions.append(
                        when(expr(rule["expectation"]), create_map())
                        .otherwise(
                            map_from_entries(
                                SparkExpectationsActions.create_rules_map(rule)
                            )
                        )
                        .alias(column)
                    )

            if len(condition_expressions) > 0:
                if rule_type in [
                    _context.get_query_dq_rule_type_name,
                    _context.get_agg_dq_rule_type_name,
                ]:
                    df = (
                        df
                        if rule_type == _context.get_agg_dq_rule_type_name
                        else _context.get_supported_df_query_dq
                    )

                    df = df.select(*condition_expressions)
                    df = df.withColumn(
                        f"meta_{rule_type}_results", array(*list(df.columns))
                    )
                    df = df.withColumn(
                        f"meta_{rule_type}_results",
                        remove_empty_maps(df[f"meta_{rule_type}_results"]),
                    ).drop(
                        *[
                            _col
                            for _col in df.columns
                            if _col != f"meta_{rule_type}_results"
                        ]
                    )
                    _context.print_dataframe_with_debugger(df)

                elif rule_type == _context.get_row_dq_rule_type_name:
                    df = df.select(col("*"), *condition_expressions)
            else:
                raise SparkExpectationsMiscException(
                    f"zero active expectations to process for {rule_type}_rules from the `dq_rules` table, "
                    f"at {f'final_{rule_type}' if _source_dq_enabled else f'final_{rule_type}' }"
                    f", please configure rules or avoid this error by setting final_{rule_type} to False"
                )

            return df

        except Exception as e:
            raise SparkExpectationsMiscException(
                f"error occurred while running expectations {e}"
            )

    @staticmethod
    def action_on_rules(
        _context: SparkExpectationsContext,
        _df_dq: DataFrame,
        _input_count: int,
        _error_count: int = 0,
        _output_count: int = 0,
        _rule_type: Optional[str] = None,
        _row_dq_flag: bool = False,
        _source_agg_dq_flag: bool = False,
        _final_agg_dq_flag: bool = False,
        _source_query_dq_flag: bool = False,
        _final_query_dq_flag: bool = False,
    ) -> DataFrame:
        """
        This function takes necessary action set by the user on the rules and returns the dataframe with results
        Args:
            _context: Provide SparkExpectationsContext
            _df_dq: Input dataframe on which data quality rules need to be applied
            _input_count: input dataset count
            _error_count: error count in the dataset
            _output_count: output count in the dataset
            _rule_type: type of rule expectations
            _row_dq_flag: Mark it as True when dq running for row level expectations
            _source_agg_dq_flag: Mark it as True when dq running for agg level expectations on source dataframe
            _final_agg_dq_flag: Mark it as True when dq running for agg level expectations on final dataframe
            _source_query_dq_flag: Mark it as True when dq running for query level expectations on source dataframe
            _final_query_dq_flag: Mark it as True when dq running for query level expectations on final dataframe
        Returns:
                DataFrame: Returns a dataframe after dropping the error from the dataset

        """
        try:
            _df_dq = _df_dq.withColumn(
                "action_if_failed", get_actions_list(col(f"meta_{_rule_type}_results"))
            ).drop(f"meta_{_rule_type}_results")

            if (
                not _df_dq.filter(
                    array_contains(_df_dq.action_if_failed, "fail")
                ).count()
                > 0
            ):
                _df_dq = _df_dq.filter(~array_contains(_df_dq.action_if_failed, "drop"))
            else:
                if _row_dq_flag:
                    _context.set_row_dq_status("Failed")
                elif _source_agg_dq_flag:
                    _context.set_source_agg_dq_status("Failed")
                elif _final_agg_dq_flag:
                    _context.set_final_agg_dq_status("Failed")
                elif _source_query_dq_flag:
                    _context.set_source_query_dq_status("Failed")
                elif _final_query_dq_flag:
                    _context.set_final_query_dq_status("Failed")

                raise SparkExpectOrFailException(
                    f"Job failed, as there is a data quality issue at {_rule_type} "
                    f"expectations and the action_if_failed "
                    "suggested to fail"
                )

            return _df_dq.drop(_df_dq.action_if_failed)

        except Exception as e:
            raise SparkExpectationsMiscException(
                f"error occured while taking action on given rules {e}"
            )



import functools
from dataclasses import dataclass
from typing import Dict, Optional, Any, Union
from pyspark import StorageLevel
from pyspark.sql import DataFrame, SparkSession
from spark_expectations import _log
from spark_expectations.config.user_config import Constants as user_config
from spark_expectations.core.context import SparkExpectationsContext
from spark_expectations.core.exceptions import (
    SparkExpectationsMiscException,
    SparkExpectationsDataframeNotReturnedException,
)
from spark_expectations.notifications.push.spark_expectations_notify import (
    SparkExpectationsNotify,
)
from spark_expectations.sinks.utils.collect_statistics import (
    SparkExpectationsCollectStatistics,
)
from spark_expectations.sinks.utils.writer import SparkExpectationsWriter
from spark_expectations.utils.actions import SparkExpectationsActions
from spark_expectations.utils.reader import SparkExpectationsReader
from spark_expectations.utils.regulate_flow import SparkExpectationsRegulateFlow


@dataclass
class SparkExpectations:
    """
    This class implements/supports running the data quality rules on a dataframe returned by a function

    Args:
        product_id: Name of the product
        rules_df: DataFrame which contains the rules. User is responsible for reading
            the rules_table in which ever system it is
        stats_table: Name of the table where the stats/audit-info need to be written
        debugger: Mark it as "True" if the debugger mode need to be enabled, by default is False
        stats_streaming_options: Provide options to override the defaults, while writing into the stats streaming table
    """

    product_id: str
    rules_df: DataFrame
    stats_table: str
    target_and_error_table_writer: "WrappedDataFrameWriter"
    stats_table_writer: "WrappedDataFrameWriter"
    debugger: bool = False
    stats_streaming_options: Optional[Dict[str, Union[str, bool]]] = None

    def __post_init__(self) -> None:
        if isinstance(self.rules_df, DataFrame):
            self.spark: SparkSession = self.rules_df.sparkSession
        else:
            raise SparkExpectationsMiscException(
                "Input rules_df is not of dataframe type"
            )
        self.actions: SparkExpectationsActions = SparkExpectationsActions()
        self._context: SparkExpectationsContext = SparkExpectationsContext(
            product_id=self.product_id, spark=self.spark
        )

        self._writer = SparkExpectationsWriter(_context=self._context)
        self._process = SparkExpectationsRegulateFlow(product_id=self.product_id)
        self._notification: SparkExpectationsNotify = SparkExpectationsNotify(
            _context=self._context
        )
        self._statistics_decorator = SparkExpectationsCollectStatistics(
            _context=self._context,
            _writer=self._writer,
        )
        self.reader = SparkExpectationsReader(_context=self._context)

        self._context.set_target_and_error_table_writer_config(
            self.target_and_error_table_writer.build()
        )
        self._context.set_stats_table_writer_config(self.stats_table_writer.build())
        self._context.set_debugger_mode(self.debugger)
        self._context.set_dq_stats_table_name(self.stats_table)
        self.rules_df = self.rules_df.persist(StorageLevel.MEMORY_AND_DISK)

    # TODO Add target_error_table_writer and stats_table_writer as parameters to this function so this takes precedence
    #  if user provides it
    def with_expectations(
        self,
        target_table: str,
        write_to_table: bool = False,
        write_to_temp_table: bool = False,
        user_conf: Optional[Dict[str, Union[str, int, bool]]] = None,
        target_table_view: Optional[str] = None,
        target_and_error_table_writer: Optional["WrappedDataFrameWriter"] = None,
    ) -> Any:
        """
        This decorator helps to wrap a function which returns dataframe and apply dataframe rules on it

        Args:
            target_table: Name of the table where the final dataframe need to be written
            write_to_table: Mark it as "True" if the dataframe need to be written as table
            write_to_temp_table: Mark it as "True" if the input dataframe need to be written to the temp table to break
                                the spark plan
            user_conf: Provide options to override the defaults, while writing into the stats streaming table
            target_table_view: This view is created after the _row_dq process to run the target agg_dq and query_dq.
                If value is not provided, defaulted to {target_table}_view
            target_and_error_table_writer: Provide the writer to write the target and error table,
                this will take precedence over the class level writer

        Returns:
            Any: Returns a function which applied the expectations on dataset
        """

        def _except(func: Any) -> Any:
            # variable used for enabling notification at different level

            _default_notification_dict: Dict[str, Union[str, int, bool]] = {
                user_config.se_notifications_on_start: False,
                user_config.se_notifications_on_completion: False,
                user_config.se_notifications_on_fail: True,
                user_config.se_notifications_on_error_drop_exceeds_threshold_breach: False,
                user_config.se_notifications_on_error_drop_threshold: 100,
            }
            _notification_dict: Dict[str, Union[str, int, bool]] = (
                {**_default_notification_dict, **user_conf}
                if user_conf
                else _default_notification_dict
            )
            _default_stats_streaming_dict: Dict[str, Union[bool, str]] = {
                user_config.se_enable_streaming: True,
                user_config.secret_type: "databricks",
                user_config.dbx_workspace_url: "https://workspace.cloud.databricks.com",
                user_config.dbx_secret_scope: "sercet_scope",
                user_config.dbx_kafka_server_url: "se_streaming_server_url_secret_key",
                user_config.dbx_secret_token_url: "se_streaming_auth_secret_token_url_key",
                user_config.dbx_secret_app_name: "se_streaming_auth_secret_appid_key",
                user_config.dbx_secret_token: "se_streaming_auth_secret_token_key",
                user_config.dbx_topic_name: "se_streaming_topic_name",
            }
            _se_stats_streaming_dict: Dict[str, Any] = (
                {**self.stats_streaming_options}
                if self.stats_streaming_options
                else _default_stats_streaming_dict
            )

            # Overwrite the writers if provided by the user in the with_expectations explicitly
            if target_and_error_table_writer:
                self._context.set_target_and_error_table_writer_config(
                    target_and_error_table_writer.build()
                )

            # need to call the get_rules_frm_table function to get the rules from the table as expectations
            expectations, rules_execution_settings = self.reader.get_rules_from_df(
                self.rules_df, target_table
            )

            _row_dq: bool = rules_execution_settings.get("row_dq", False)
            _source_agg_dq: bool = rules_execution_settings.get("source_agg_dq", False)
            _target_agg_dq: bool = rules_execution_settings.get("target_agg_dq", False)
            _source_query_dq: bool = rules_execution_settings.get(
                "source_query_dq", False
            )
            _target_query_dq: bool = rules_execution_settings.get(
                "target_query_dq", False
            )
            _target_table_view: str = (
                target_table_view
                if target_table_view
                else f"{target_table.split('.')[-1]}_view"
            )

            _notification_on_start: bool = (
                bool(_notification_dict[user_config.se_notifications_on_start])
                if isinstance(
                    _notification_dict[user_config.se_notifications_on_start],
                    bool,
                )
                else False
            )
            _notification_on_completion: bool = (
                bool(_notification_dict[user_config.se_notifications_on_completion])
                if isinstance(
                    _notification_dict[user_config.se_notifications_on_completion],
                    bool,
                )
                else False
            )
            _notification_on_fail: bool = (
                bool(_notification_dict[user_config.se_notifications_on_fail])
                if isinstance(
                    _notification_dict[user_config.se_notifications_on_fail],
                    bool,
                )
                else False
            )
            _notification_on_error_drop_exceeds_threshold_breach: bool = (
                bool(
                    _notification_dict[
                        user_config.se_notifications_on_error_drop_exceeds_threshold_breach
                    ]
                )
                if isinstance(
                    _notification_dict[
                        user_config.se_notifications_on_error_drop_exceeds_threshold_breach
                    ],
                    bool,
                )
                else False
            )
            _error_drop_threshold: int = (
                int(
                    _notification_dict[
                        user_config.se_notifications_on_error_drop_threshold
                    ]
                )
                if isinstance(
                    _notification_dict[
                        user_config.se_notifications_on_error_drop_threshold
                    ],
                    int,
                )
                else 100
            )

            self.reader.set_notification_param(user_conf)
            self._context.set_notification_on_start(_notification_on_start)
            self._context.set_notification_on_completion(_notification_on_completion)
            self._context.set_notification_on_fail(_notification_on_fail)

            self._context.set_se_streaming_stats_dict(_se_stats_streaming_dict)

            @self._notification.send_notification_decorator
            @self._statistics_decorator.collect_stats_decorator
            @functools.wraps(func)
            def wrapper(*args: tuple, **kwargs: dict) -> DataFrame:
                try:
                    _log.info("The function dataframe is getting created")
                    # _df: DataFrame = func(*args, **kwargs)
                    _df: DataFrame = func(*args, **kwargs)
                    table_name: str = self._context.get_table_name

                    _input_count = _df.count()
                    _output_count: int = 0
                    _error_count: int = 0
                    _source_dq_df: Optional[DataFrame] = None
                    _source_query_dq_df: Optional[DataFrame] = None
                    _row_dq_df: DataFrame = _df
                    _final_dq_df: Optional[DataFrame] = None
                    _final_query_dq_df: Optional[DataFrame] = None

                    # initialize variable with default values through set
                    self._context.set_dq_run_status()
                    self._context.set_source_agg_dq_status()
                    self._context.set_source_query_dq_status()
                    self._context.set_row_dq_status()
                    self._context.set_final_agg_dq_status()
                    self._context.set_final_query_dq_status()
                    self._context.set_input_count()
                    self._context.set_error_count()
                    self._context.set_output_count()
                    self._context.set_source_agg_dq_result()
                    self._context.set_final_agg_dq_result()
                    self._context.set_source_query_dq_result()
                    self._context.set_final_query_dq_result()
                    self._context.set_summarised_row_dq_res()

                    # initialize variables of start and end time with default values
                    self._context._source_agg_dq_start_time = None
                    self._context._final_agg_dq_start_time = None
                    self._context._source_query_dq_start_time = None
                    self._context._final_query_dq_start_time = None
                    self._context._row_dq_start_time = None

                    self._context._source_agg_dq_end_time = None
                    self._context._final_agg_dq_end_time = None
                    self._context._source_query_dq_end_time = None
                    self._context._final_query_dq_end_time = None
                    self._context._row_dq_end_time = None

                    self._context.set_input_count(_input_count)
                    self._context.set_error_drop_threshold(_error_drop_threshold)

                    if isinstance(_df, DataFrame):
                        _log.info("The function dataframe is created")
                        self._context.set_table_name(table_name)
                        if write_to_temp_table:
                            _log.info("Dropping to temp table started")
                            self.spark.sql(f"drop table if exists {table_name}_temp")
                            _log.info("Dropping to temp table completed")
                            _log.info("Writing to temp table started")
                            self._writer.save_df_as_table(
                                _df,
                                f"{table_name}_temp",
                                self._context.get_target_and_error_table_writer_config,
                            )
                            _log.info("Read from temp table started")
                            _df = self.spark.sql(f"select * from {table_name}_temp")
                            _log.info("Read from temp table completed")

                        func_process = self._process.execute_dq_process(
                            _context=self._context,
                            _actions=self.actions,
                            _writer=self._writer,
                            _notification=self._notification,
                            expectations=expectations,
                            table_name=table_name,
                            _input_count=_input_count,
                        )

                        if _source_agg_dq is True:
                            _log.info(
                                "started processing data quality rules for agg level expectations on soure dataframe"
                            )
                            self._context.set_source_agg_dq_status("Failed")
                            self._context.set_source_agg_dq_start_time()
                            # In this steps source agg data quality expectations runs on raw_data
                            # returns:
                            #        _source_dq_df: applied data quality dataframe,
                            #        _dq_source_agg_results: source aggregation result in dictionary
                            #        _: place holder for error data at row level
                            #        status: status of the execution

                            (
                                _source_dq_df,
                                _dq_source_agg_results,
                                _,
                                status,
                            ) = func_process(
                                _df,
                                self._context.get_agg_dq_rule_type_name,
                                source_agg_dq_flag=True,
                            )
                            self._context.set_source_agg_dq_status(status)
                            self._context.set_source_agg_dq_end_time()

                            _log.info(
                                "ended processing data quality rules for agg level expectations on source dataframe"
                            )

                        if _source_query_dq is True:
                            _log.info(
                                "started processing data quality rules for query level expectations on soure dataframe"
                            )
                            self._context.set_source_query_dq_status("Failed")
                            self._context.set_source_query_dq_start_time()
                            # In this steps source query data quality expectations runs on raw_data
                            # returns:
                            #        _source_query_dq_df: applied data quality dataframe,
                            #        _dq_source_query_results: source query dq results in dictionary
                            #        _: place holder for error data at row level
                            #        status: status of the execution

                            (
                                _source_query_dq_df,
                                _dq_source_query_results,
                                _,
                                status,
                            ) = func_process(
                                _df,
                                self._context.get_query_dq_rule_type_name,
                                source_query_dq_flag=True,
                            )
                            self._context.set_source_query_dq_status(status)
                            self._context.set_source_query_dq_end_time()
                            _log.info(
                                "ended processing data quality rules for query level expectations on source dataframe"
                            )

                        if _row_dq is True:
                            _log.info(
                                "started processing data quality rules for row level expectations"
                            )
                            self._context.set_row_dq_status("Failed")
                            self._context.set_row_dq_start_time()
                            # In this steps row level data quality expectations runs on raw_data
                            # returns:
                            #        _row_dq_df: applied data quality dataframe at row level on raw dataframe,
                            #        _: place holder for aggregation
                            #        _error_count: number of error records
                            #        status: status of the execution
                            (_row_dq_df, _, _error_count, status) = func_process(
                                _df,
                                self._context.get_row_dq_rule_type_name,
                                row_dq_flag=True,
                            )
                            self._context.set_error_count(_error_count)

                            _row_dq_df.createOrReplaceTempView(_target_table_view)

                            _output_count = _row_dq_df.count() if _row_dq_df else 0
                            self._context.set_output_count(_output_count)

                            self._context.set_row_dq_status(status)
                            self._context.set_row_dq_end_time()

                            if (
                                _notification_on_error_drop_exceeds_threshold_breach
                                is True
                                and (100 - self._context.get_output_percentage)
                                >= _error_drop_threshold
                            ):
                                self._notification.notify_on_exceeds_of_error_threshold()
                                # raise SparkExpectationsErrorThresholdExceedsException(
                                #     "An error has taken place because"
                                #     " the set limit for acceptable"
                                #     " errors, known as the error"
                                #     " threshold, has been surpassed"
                                # )
                            _log.info(
                                "ended processing data quality rules for row level expectations"
                            )

                        if _row_dq is True and _target_agg_dq is True:
                            _log.info(
                                "started processing data quality rules for agg level expectations on final dataframe"
                            )
                            self._context.set_final_agg_dq_status("Failed")
                            self._context.set_final_agg_dq_start_time()
                            # In this steps final agg data quality expectations run on final dataframe
                            # returns:
                            #        _final_dq_df: applied data quality dataframe at row level on raw dataframe,
                            #        _dq_final_agg_results: final agg dq result in dictionary
                            #        _: number of error records
                            #        status: status of the execution
                            (
                                _final_dq_df,
                                _dq_final_agg_results,
                                _,
                                status,
                            ) = func_process(
                                _row_dq_df,
                                self._context.get_agg_dq_rule_type_name,
                                final_agg_dq_flag=True,
                                error_count=_error_count,
                                output_count=_output_count,
                            )
                            self._context.set_final_agg_dq_status(status)
                            self._context.set_final_agg_dq_end_time()
                            _log.info(
                                "ended processing data quality rules for agg level expectations on final dataframe"
                            )

                        if _row_dq is True and _target_query_dq is True:
                            _log.info(
                                "started processing data quality rules for query level expectations on final dataframe"
                            )
                            self._context.set_final_query_dq_status("Failed")
                            self._context.set_final_query_dq_start_time()
                            # In this steps final query dq data quality expectations run on final dataframe
                            # returns:
                            #        _final_query_dq_df: applied data quality dataframe at row level on raw dataframe,
                            #        _dq_final_query_results: final query dq result in dictionary
                            #        _: number of error records
                            #        status: status of the execution

                            _row_dq_df.createOrReplaceTempView(_target_table_view)

                            (
                                _final_query_dq_df,
                                _dq_final_query_results,
                                _,
                                status,
                            ) = func_process(
                                _row_dq_df,
                                self._context.get_query_dq_rule_type_name,
                                final_query_dq_flag=True,
                                error_count=_error_count,
                                output_count=_output_count,
                            )
                            self._context.set_final_query_dq_status(status)
                            self._context.set_final_query_dq_end_time()

                            _log.info(
                                "ended processing data quality rules for query level expectations on final dataframe"
                            )

                        # TODO if row_dq is False and source_agg/source_query is True then we need to write the
                        #  dataframe into the target table
                        if write_to_table:
                            _log.info("Writing into the final table started")
                            self._writer.save_df_as_table(
                                _row_dq_df,
                                f"{table_name}",
                                self._context.get_target_and_error_table_writer_config,
                            )
                            _log.info("Writing into the final table ended")

                    else:
                        raise SparkExpectationsDataframeNotReturnedException(
                            "error occurred while processing spark "
                            "expectations due to given dataframe is not type of dataframe"
                        )
                    self.spark.catalog.clearCache()

                    return _row_dq_df

                except Exception as e:
                    raise SparkExpectationsMiscException(
                        f"error occurred while processing spark expectations {e}"
                    )

            return wrapper

        return _except


class WrappedDataFrameWriter:
    """
    A builder pattern class that mimics the functions of PySpark's DataFrameWriter.

    This class allows for chaining methods to set configurations like mode, format, partitioning,
    options, and bucketing. It does not require a DataFrame object and is designed purely to
    collect and return configurations.

    Example usage:
    --------------
    writer = WrappedDataFrameWriter().mode("overwrite")\
                                   .format("parquet")\
                                   .partitionBy("date", "region")\
                                   .option("compression", "gzip")\
                                   .options(path="/path/to/output", inferSchema="true")\
                                   .bucketBy(4, "country", "city")

    config = writer.build()
    print(config)

    Attributes:
    -----------
    _mode : str
        The mode for writing (e.g., "overwrite", "append").
    _format : str
        The format for writing (e.g., "parquet", "csv").
    _partition_by : list
        Columns by which the data should be partitioned.
    _options : dict
        Additional options for writing.
    _bucket_by : dict
        Configuration for bucketing, including number of buckets and columns.
    """

    def __init__(self) -> None:
        self._mode: Optional[str] = None
        self._format: Optional[str] = None
        self._partition_by: list = []
        self._options: dict[str, str] = {}
        self._bucket_by: Dict[str, Union[int, tuple]] = {}
        self._sort_by: list = []

    def mode(self, saveMode: str) -> "WrappedDataFrameWriter":  # noqa: N803
        """Set the mode for writing."""
        self._mode = saveMode
        return self

    def format(self, source: str) -> "WrappedDataFrameWriter":
        """Set the format for writing."""
        self._format = source
        return self

    def partitionBy(self, *columns: str) -> "WrappedDataFrameWriter":  # noqa: N802
        """Set the columns by which the data should be partitioned."""
        self._partition_by.extend(columns)
        return self

    def option(self, key: str, value: str) -> "WrappedDataFrameWriter":
        """Set a single option for writing."""
        self._options[key] = value
        return self

    def options(self, **options: str) -> "WrappedDataFrameWriter":
        """Set multiple options for writing."""
        self._options.update(options)
        return self

    def bucketBy(  # noqa: N802
        self, num_buckets: int, *columns: str
    ) -> "WrappedDataFrameWriter":
        """Set the configuration for bucketing."""
        self._bucket_by["num_buckets"] = num_buckets
        self._bucket_by["columns"] = columns
        return self

    def sortBy(self, *columns: str) -> "WrappedDataFrameWriter":  # noqa: N802
        """Set the configuration for bucketing."""
        self._sort_by.extend(columns)
        return self

    def build(self) -> Dict[str, Union[str, list, dict, tuple, int, None]]:
        """Return the collected configurations."""
        if self._format is not None and self._format.lower() == "delta":
            if self._bucket_by is not None and self._bucket_by:
                raise SparkExpectationsMiscException(
                    "Bucketing is not supported for delta tables yet..."
                )

        return {
            "mode": self._mode,
            "format": self._format,
            "partitionBy": self._partition_by,
            "options": self._options,
            "bucketBy": self._bucket_by,
            "sortBy": self._sort_by,
        }

        # config = {}
        #
        # if cls._mode:
        #     config["mode"] = cls._mode
        # if cls._format:
        #     config["format"] = cls._format
        # if cls._partition_by:
        #     config["partitionBy"] = cls._partition_by
        # if cls._options:
        #     config["options"] = cls._options
        # if cls._bucket_by:
        #     config["bucketBy"] = cls._bucket_by
        # if cls._sort_by:
        #     config["sortBy"] = cls._sort_by
        #
        # return config



from dataclasses import dataclass


@dataclass
class Constants:
    # declare const user config variables for email notification
    se_notifications_enable_email = "spark.expectations.notifications.email.enabled"
    se_notifications_email_smtp_host = (
        "spark.expectations.notifications.email.smtp_host"
    )
    se_notifications_email_smtp_port = (
        "spark.expectations.notifications.email.smtp_port"
    )
    se_notifications_email_from = "spark.expectations.notifications.email.from"
    se_notifications_email_to_other_mail_id = (
        "spark.expectations.notifications.email.to.other.mail.com"
    )
    se_notifications_email_subject = "spark.expectations.notifications.email.subject"

    # declare const user config variables for slack notification
    se_notifications_enable_slack = "spark.expectations.notifications.slack.enabled"
    se_notifications_slack_webhook_url = (
        "spark.expectations.notifications.slack.webhook_url"
    )

    # declare const user config variables for teams notification
    se_notifications_enable_teams = "spark.expectations.notifications.teams.enabled"
    se_notifications_teams_webhook_url = (
        "spark.expectations.notifications.teams.webhook_url"
    )

    se_notifications_on_start = "spark.expectations.notifications.on_start"
    se_notifications_on_completion = "spark.expectations.notifications.on.completion"
    se_notifications_on_fail = "spark.expectations.notifications.on.fail"
    se_notifications_on_error_drop_exceeds_threshold_breach = (
        "spark.expectations.notifications.on.error.drop.exceeds.threshold.breach"
    )
    se_notifications_on_error_drop_threshold = (
        "spark.expectations.notifications.error.drop.threshold"
    )

    se_enable_streaming = "se.enable.streaming"

    secret_type = "se.streaming.secret.type"

    cbs_url = "se.streaming.cerberus.url"
    cbs_sdb_path = "se.streaming.cerberus.sdb.path"
    cbs_kafka_server_url = "se.streaming.cerberus.kafka.server.url"
    cbs_secret_token_url = "se.streaming.cbs.secret.token.url"
    cbs_secret_app_name = "se.streaming.cbs.secret.app.name"
    cbs_secret_token = "se.streaming.cerberus.secret.token"
    cbs_topic_name = "se.streaming.cerberus.token.name"

    dbx_workspace_url = "se.streaming.dbx.workspace.url"
    dbx_secret_scope = "se.streaming.dbx.secret.scope"
    dbx_kafka_server_url = "se.streaming.dbx.kafka.server.url"
    dbx_secret_token_url = "se.streaming.dbx.secret.token.url"
    dbx_secret_app_name = "se.streaming.dbx.secret.app.name"
    dbx_secret_token = "se.streaming.dbx.secret.token"
    dbx_topic_name = "se.streaming.dbx.topic.name"



import os
from pyspark.sql.session import SparkSession

os.environ["SPARKEXPECTATIONS_ENV"] = "local"

CURRENT_DIR = os.path.dirname(os.path.abspath(__file__))

RULES_TABLE_SCHEMA = """ ( product_id STRING,
    table_name STRING,
    rule_type STRING,
    rule STRING,
    column_name STRING,
    expectation STRING,
    action_if_failed STRING,
    tag STRING,
    description STRING,
    enable_for_source_dq_validation BOOLEAN, 
    enable_for_target_dq_validation BOOLEAN,
    is_active BOOLEAN,
    enable_error_drop_alert BOOLEAN,
    error_drop_threshold INT )
"""

RULES_DATA = """ 
    ("your_product", "dq_spark_local.customer_order",  "row_dq", "customer_id_is_not_null", "customer_id", "customer_id is not null","drop", "validity", "customer_id ishould not be null", true, true,false, false, 0)
    ,("your_product", "dq_spark_local.customer_order", "row_dq", "sales_greater_than_zero", "sales", "sales > 2", "drop", "accuracy", "sales value should be greater than zero", true, true, true, false, 0)
    ,("your_product", "dq_spark_local.customer_order", "row_dq", "discount_threshold", "discount", "discount*100 < 60","drop", "validity", "discount should be less than 40", true, true, false, false, 0)
    ,("your_product", "dq_spark_local.customer_order", "row_dq", "ship_mode_in_set", "ship_mode", "lower(trim(ship_mode)) in('second class', 'standard class', 'standard class')", "drop", "validity", "ship_mode mode belongs in the sets", true, true, false, false, 0)
    ,("your_product", "dq_spark_local.customer_order", "row_dq", "profit_threshold", "profit", "profit>0", "drop", "validity", "profit threshold should be greater tahn 0", true, true, false, true, 0)
    
    ,("your_product", "dq_spark_local.customer_order", "agg_dq", "sum_of_sales", "sales", "sum(sales)>10000", "ignore", "validity", "regex format validation for quantity",  true, true, true, false, 0)
    ,("your_product", "dq_spark_local.customer_order", "agg_dq", "sum_of_quantity", "quantity", "sum(sales)>10000", "ignore", "validity", "regex format validation for quantity", true, true, true, false, 0)
    ,("your_product", "dq_spark_local.customer_order", "agg_dq", "distinct_of_ship_mode", "ship_mode", "count(distinct ship_mode)<=3", "ignore", "validity", "regex format validation for quantity", true, true, true, false, 0)
    ,("your_product", "dq_spark_local.customer_order", "agg_dq", "row_count", "*", "count(*)>=10000", "ignore", "validity", "regex format validation for quantity", true, true, true, false, 0)

    ,("your_product", "dq_spark_local.customer_order", "query_dq", "product_missing_count_threshold", "*", "((select count(distinct product_id) from product) - (select count(distinct product_id) from order))>(select count(distinct product_id) from product)*0.2", "ignore", "validity", "row count threshold", true, true, true, false, 0)
    ,("your_product", "dq_spark_local.customer_order", "query_dq", "product_category", "*", "(select count(distinct category) from product) < 5", "ignore", "validity", "distinct product category", true, true, true, false, 0)
    ,("your_product", "dq_spark_local.customer_order", "query_dq", "row_count_in_order", "*", "(select count(*) from order)<10000", "ignore", "accuracy", "count of the row in order dataset", true, true, true, false, 0)
    
"""


def set_up_kafka() -> None:
    print("create or run if exist docker container")
    os.system(f"sh {CURRENT_DIR}/docker_scripts/docker_kafka_start_script.sh")


def add_kafka_jars(builder: SparkSession.builder) -> SparkSession.builder:
    return builder.config(  # below jars are used only in the local env, not coupled with databricks or EMR
        "spark.jars",
        f"{CURRENT_DIR}/../../jars/spark-sql-kafka-0-10_2.12-3.0.0.jar,"
        f"{CURRENT_DIR}/../../jars/kafka-clients-3.0.0.jar,"
        f"{CURRENT_DIR}/../../jars/commons-pool2-2.8.0.jar,"
        f"{CURRENT_DIR}/../../jars/spark-token-provider-kafka-0-10_2.12-3.0.0.jar",
    )


def set_up_iceberg() -> SparkSession:
    set_up_kafka()
    spark = add_kafka_jars(
        SparkSession.builder.config(
            "spark.jars.packages",
            "org.apache.iceberg:iceberg-spark-runtime-3.4_2.12:1.3.1",
        )
        .config(
            "spark.sql.extensions",
            "org.apache.iceberg.spark.extensions.IcebergSparkSessionExtensions",
        )
        .config(
            "spark.sql.catalog.spark_catalog",
            "org.apache.iceberg.spark.SparkSessionCatalog",
        )
        .config("spark.sql.catalog.spark_catalog.type", "hadoop")
        .config("spark.sql.catalog.spark_catalog.warehouse", "/tmp/hive/warehouse")
        .config("spark.sql.catalog.local", "org.apache.iceberg.spark.SparkCatalog")
        .config("spark.sql.catalog.local.type", "hadoop")
        .config("spark.sql.catalog.local.warehouse", "/tmp/hive/warehouse")
    ).getOrCreate()

    os.system("rm -rf /tmp/hive/warehouse/dq_spark_local")

    spark.sql("create database if not exists spark_catalog.dq_spark_local")
    spark.sql(" use spark_catalog.dq_spark_local")

    spark.sql("drop table if exists dq_spark_local.dq_stats")

    spark.sql("drop table if exists dq_spark_local.dq_rules")

    spark.sql(
        f" CREATE TABLE dq_spark_local.dq_rules {RULES_TABLE_SCHEMA} USING ICEBERG"
    )
    spark.sql(f" INSERT INTO dq_spark_local.dq_rules  values {RULES_DATA} ")

    spark.sql("select * from dq_spark_local.dq_rules").show(truncate=False)
    return spark


def set_up_bigquery(materialization_dataset: str) -> SparkSession:
    set_up_kafka()
    spark = add_kafka_jars(
        SparkSession.builder.config(
            "spark.jars.packages",
            "com.google.cloud.spark:spark-bigquery-with-dependencies_2.12:0.30.0",
        )
    ).getOrCreate()
    spark._jsc.hadoopConfiguration().set(
        "fs.gs.impl", "com.google.cloud.hadoop.fs.gcs.GoogleHadoopFileSystem"
    )
    spark.conf.set("viewsEnabled", "true")
    spark.conf.set("materializationDataset", materialization_dataset)

    # Add dependencies like gcs-connector-hadoop3-2.2.6-SNAPSHOT-shaded.jar, spark-avro_2.12-3.4.1.jar if you wanted to use indirect method for reading/writing
    return spark


def set_up_delta() -> SparkSession:
    set_up_kafka()

    builder = add_kafka_jars(
        SparkSession.builder.config(
            "spark.sql.extensions", "io.delta.sql.DeltaSparkSessionExtension"
        )
        .config("spark.jars.packages", "io.delta:delta-core_2.12:2.4.0")
        .config(
            "spark.sql.catalog.spark_catalog",
            "org.apache.spark.sql.delta.catalog.DeltaCatalog",
        )
        .config("spark.sql.warehouse.dir", "/tmp/hive/warehouse")
        .config("spark.driver.extraJavaOptions", "-Dderby.system.home=/tmp/derby")
        .config("spark.jars.ivy", "/tmp/ivy2")
    )
    spark = builder.getOrCreate()

    os.system("rm -rf /tmp/hive/warehouse/dq_spark_local.db")

    spark.sql("create database if not exists dq_spark_local")
    spark.sql("use dq_spark_local")

    spark.sql("drop table if exists dq_stats")

    spark.sql("drop table if exists dq_rules")

    spark.sql(f" CREATE TABLE dq_rules {RULES_TABLE_SCHEMA} USING DELTA")
    spark.sql(f" INSERT INTO dq_rules  values {RULES_DATA}")

    spark.sql("select * from dq_rules").show(truncate=False)
    return spark



from typing import Dict, Union
from pyspark.sql import DataFrame
from spark_expectations.sinks.plugins.base_writer import (
    SparkExpectationsSinkWriter,
    spark_expectations_writer_impl,
)
from spark_expectations.core.exceptions import SparkExpectationsMiscException
from spark_expectations import _log


class SparkExpectationsKafkaWritePluginImpl(SparkExpectationsSinkWriter):
    """
    class helps to write the stats data into the NSP
    """

    @spark_expectations_writer_impl
    def writer(
        self, _write_args: Dict[Union[str], Union[str, bool, Dict[str, str], DataFrame]]
    ) -> None:
        """
        The functions helps to write data into the kafka topic
        Args:
            _write_args:

        Returns:

        """

        try:
            # kafka_options = {
            #     "kafka.bootstrap.servers": "localhost:9092",
            #     "topic": _write_args.get("nsp_topic_name"),
            #     "failOnDataLoss": "true",
            # }

            if _write_args.pop("enable_se_streaming"):
                _log.info("started write stats data into kafka stats topic")

                df: DataFrame = _write_args.get("stats_df")

                df.selectExpr("to_json(struct(*)) AS value").write.format("kafka").mode(
                    "append"
                ).options(**_write_args.get("kafka_write_options")).save()

                _log.info("ended writing stats data into kafka stats topic")

        except Exception as e:
            raise SparkExpectationsMiscException(
                f"error occurred while saving data into kafka {e}"
            )
